# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from typing import Any, TypedDict

from mautrix.types import (
    Member,
    Membership,
    MemberStateEventContent,
    PowerLevelStateEventContent,
    RoomEncryptionStateEventContent,
    RoomID,
    UserID,
)

from .abstract import StateStore


class SerializedStateStore(TypedDict):
    members: dict[RoomID, dict[UserID, Any]]
    full_member_list: dict[RoomID, bool]
    power_levels: dict[RoomID, Any]
    encryption: dict[RoomID, Any]


class MemoryStateStore(StateStore):
    members: dict[RoomID, dict[UserID, Member]]
    full_member_list: dict[RoomID, bool]
    power_levels: dict[RoomID, PowerLevelStateEventContent]
    encryption: dict[RoomID, RoomEncryptionStateEventContent | None]

    def __init__(self) -> None:
        self.members = {}
        self.full_member_list = {}
        self.power_levels = {}
        self.encryption = {}

    def serialize(self) -> SerializedStateStore:
        """
        Convert the data in the store into a JSON-friendly dict.

        Returns: A dict that can be safely serialized with most object serialization methods.
        """
        return {
            "members": {
                room_id: {user_id: member.serialize() for user_id, member in members.items()}
                for room_id, members in self.members.items()
            },
            "full_member_list": self.full_member_list,
            "power_levels": {
                room_id: content.serialize() for room_id, content in self.power_levels.items()
            },
            "encryption": {
                room_id: (content.serialize() if content is not None else None)
                for room_id, content in self.encryption.items()
            },
        }

    def deserialize(self, data: SerializedStateStore) -> None:
        """
        Parse a previously serialized dict into this state store.

        Args:
            data: A dict returned by :meth:`serialize`.
        """
        self.members = {
            room_id: {user_id: Member.deserialize(member) for user_id, member in members.items()}
            for room_id, members in data["members"].items()
        }
        self.full_member_list = data["full_member_list"]
        self.power_levels = {
            room_id: PowerLevelStateEventContent.deserialize(content)
            for room_id, content in data["power_levels"].items()
        }
        self.encryption = {
            room_id: (
                RoomEncryptionStateEventContent.deserialize(content)
                if content is not None
                else None
            )
            for room_id, content in data["encryption"].items()
        }

    async def get_member(self, room_id: RoomID, user_id: UserID) -> Member | None:
        try:
            return self.members[room_id][user_id]
        except KeyError:
            return None

    async def set_member(
        self, room_id: RoomID, user_id: UserID, member: Member | MemberStateEventContent
    ) -> None:
        if not isinstance(member, Member):
            member = Member(
                membership=member.membership,
                avatar_url=member.avatar_url,
                displayname=member.displayname,
            )
        try:
            self.members[room_id][user_id] = member
        except KeyError:
            self.members[room_id] = {user_id: member}

    async def set_membership(
        self, room_id: RoomID, user_id: UserID, membership: Membership
    ) -> None:
        try:
            room_members = self.members[room_id]
        except KeyError:
            self.members[room_id] = {user_id: Member(membership=membership)}
            return
        try:
            room_members[user_id].membership = membership
        except (KeyError, TypeError):
            room_members[user_id] = Member(membership=membership)

    async def get_member_profiles(
        self,
        room_id: RoomID,
        memberships: tuple[Membership, ...] = (Membership.JOIN, Membership.INVITE),
    ) -> dict[UserID, Member]:
        try:
            return {
                user_id: member
                for user_id, member in self.members[room_id].items()
                if member.membership in memberships
            }
        except KeyError:
            return {}

    async def set_members(
        self,
        room_id: RoomID,
        members: dict[UserID, Member | MemberStateEventContent],
        only_membership: Membership | None = None,
    ) -> None:
        old_members = {}
        if only_membership is not None:
            old_members = {
                user_id: member
                for user_id, member in self.members.get(room_id, {}).items()
                if member.membership != only_membership
            }
        self.members[room_id] = {
            user_id: (
                member
                if isinstance(member, Member)
                else Member(
                    membership=member.membership,
                    avatar_url=member.avatar_url,
                    displayname=member.displayname,
                )
            )
            for user_id, member in members.items()
        }
        self.members[room_id].update(old_members)
        self.full_member_list[room_id] = True

    async def has_full_member_list(self, room_id: RoomID) -> bool:
        return self.full_member_list.get(room_id, False)

    async def has_power_levels_cached(self, room_id: RoomID) -> bool:
        return room_id in self.power_levels

    async def get_power_levels(self, room_id: RoomID) -> PowerLevelStateEventContent | None:
        return self.power_levels.get(room_id)

    async def set_power_levels(
        self, room_id: RoomID, content: PowerLevelStateEventContent
    ) -> None:
        self.power_levels[room_id] = content

    async def has_encryption_info_cached(self, room_id: RoomID) -> bool:
        return room_id in self.encryption

    async def is_encrypted(self, room_id: RoomID) -> bool | None:
        try:
            return self.encryption[room_id] is not None
        except KeyError:
            return None

    async def get_encryption_info(self, room_id: RoomID) -> RoomEncryptionStateEventContent:
        return self.encryption.get(room_id)

    async def set_encryption_info(
        self, room_id: RoomID, content: RoomEncryptionStateEventContent | dict[str, Any]
    ) -> None:
        if not isinstance(content, RoomEncryptionStateEventContent):
            content = RoomEncryptionStateEventContent.deserialize(content)
        self.encryption[room_id] = content



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from mautrix.types import EventID

from .handler import SECTION_RELAY, CommandEvent, command_handler


@command_handler(
    needs_auth=True,
    management_only=False,
    name="set-relay",
    help_section=SECTION_RELAY,
    help_text="Relay messages in this room through your account.",
    is_enabled_for=lambda evt: evt.config["bridge.relay.enabled"],
)
async def set_relay(evt: CommandEvent) -> EventID:
    if not evt.is_portal:
        return await evt.reply("This is not a portal room.")
    await evt.portal.set_relay_user(evt.sender)
    return await evt.reply(
        "Messages from non-logged-in users in this room will now be bridged "
        "through your account."
    )


@command_handler(
    needs_auth=True,
    management_only=False,
    name="unset-relay",
    help_section=SECTION_RELAY,
    help_text="Stop relaying messages in this room.",
    is_enabled_for=lambda evt: evt.config["bridge.relay.enabled"],
)
async def unset_relay(evt: CommandEvent) -> EventID:
    if not evt.is_portal:
        return await evt.reply("This is not a portal room.")
    elif not evt.portal.has_relay:
        return await evt.reply("This room does not have a relay user set.")
    await evt.portal.set_relay_user(None)
    return await evt.reply("Messages from non-logged-in users will no longer be bridged.")



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from typing import IO, Any
from pathlib import Path

from mautrix.types import (
    Member,
    Membership,
    MemberStateEventContent,
    PowerLevelStateEventContent,
    RoomEncryptionStateEventContent,
    RoomID,
    UserID,
)
from mautrix.util.file_store import Filer, FileStore

from .memory import MemoryStateStore


class FileStateStore(MemoryStateStore, FileStore):
    def __init__(
        self,
        path: str | Path | IO,
        filer: Filer | None = None,
        binary: bool = True,
        save_interval: float = 60.0,
    ) -> None:
        FileStore.__init__(self, path, filer, binary, save_interval)
        MemoryStateStore.__init__(self)

    async def set_membership(
        self, room_id: RoomID, user_id: UserID, membership: Membership
    ) -> None:
        await super().set_membership(room_id, user_id, membership)
        self._time_limited_flush()

    async def set_member(
        self, room_id: RoomID, user_id: UserID, member: Member | MemberStateEventContent
    ) -> None:
        await super().set_member(room_id, user_id, member)
        self._time_limited_flush()

    async def set_members(
        self,
        room_id: RoomID,
        members: dict[UserID, Member | MemberStateEventContent],
        only_membership: Membership | None = None,
    ) -> None:
        await super().set_members(room_id, members, only_membership)
        self._time_limited_flush()

    async def set_encryption_info(
        self, room_id: RoomID, content: RoomEncryptionStateEventContent | dict[str, Any]
    ) -> None:
        await super().set_encryption_info(room_id, content)
        self._time_limited_flush()

    async def set_power_levels(
        self, room_id: RoomID, content: PowerLevelStateEventContent
    ) -> None:
        await super().set_power_levels(room_id, content)
        self._time_limited_flush()



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from typing import Iterable

from sqlalchemy import Column, Enum, Text

from mautrix.types import ContentURI, Member, Membership, RoomID, UserID
from mautrix.util.db import Base

from .mx_room_state import RoomState


class UserProfile(Base):
    __tablename__ = "mx_user_profile"

    room_id: RoomID = Column(Text, primary_key=True)
    user_id: UserID = Column(Text, primary_key=True)
    membership: Membership = Column(Enum(Membership), nullable=False, default=Membership.LEAVE)
    displayname: str = Column(Text, nullable=True)
    avatar_url: ContentURI = Column(Text, nullable=True)

    def member(self) -> Member:
        return Member(
            membership=self.membership, displayname=self.displayname, avatar_url=self.avatar_url
        )

    @classmethod
    def get(cls, room_id: RoomID, user_id: UserID) -> UserProfile | None:
        return cls._select_one_or_none((cls.c.room_id == room_id) & (cls.c.user_id == user_id))

    @classmethod
    def all_in_room(
        cls,
        room_id: RoomID,
        memberships: tuple[Membership, ...],
        prefix: str = None,
        suffix: str = None,
        bot: str = None,
    ) -> Iterable[UserProfile]:
        clause = cls.c.membership == memberships[0]
        for membership in memberships[1:]:
            clause |= cls.c.membership == membership
        clause &= cls.c.room_id == room_id
        if bot:
            clause &= cls.c.user_id != bot
        if prefix:
            clause &= ~cls.c.user_id.startswith(prefix, autoescape=True)
        if suffix:
            clause &= ~cls.c.user_id.startswith(suffix, autoescape=True)
        return cls._select_all(clause)

    @classmethod
    def find_rooms_with_user(cls, user_id: UserID) -> Iterable[UserProfile]:
        return cls._select_all(
            (cls.c.user_id == user_id)
            & (cls.c.room_id == RoomState.c.room_id)
            & (RoomState.c.is_encrypted == True)
        )

    @classmethod
    def delete_all(cls, room_id: RoomID) -> None:
        with cls.db.begin() as conn:
            conn.execute(cls.t.delete().where(cls.c.room_id == room_id))

    @classmethod
    def bulk_replace(
        cls,
        room_id: RoomID,
        members: dict[UserID, Member],
        only_membership: Membership | None = None,
    ) -> None:
        with cls.db.begin() as conn:
            delete_condition = cls.c.room_id == room_id
            if only_membership is not None:
                delete_condition &= cls.c.membership == only_membership
            cls.db.execute(cls.t.delete().where(delete_condition))
            conn.execute(
                cls.t.insert(),
                [
                    dict(
                        room_id=room_id,
                        user_id=user_id,
                        membership=member.membership,
                        displayname=member.displayname,
                        avatar_url=member.avatar_url,
                    )
                    for user_id, member in members.items()
                ],
            )



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from typing import Any, Awaitable, Iterable, TypeVar
from urllib.parse import quote as urllib_quote

from mautrix.api import Method, Path
from mautrix.client import ClientAPI, StoreUpdatingAPI
from mautrix.errors import (
    IntentError,
    MAlreadyJoined,
    MatrixRequestError,
    MBadState,
    MForbidden,
    MNotFound,
    MUserInUse,
)
from mautrix.types import (
    JSON,
    BatchID,
    BatchSendEvent,
    BatchSendResponse,
    BatchSendStateEvent,
    ContentURI,
    EventContent,
    EventID,
    EventType,
    JoinRule,
    JoinRulesStateEventContent,
    Member,
    Membership,
    PowerLevelStateEventContent,
    PresenceState,
    RoomAvatarStateEventContent,
    RoomID,
    RoomNameStateEventContent,
    RoomPinnedEventsStateEventContent,
    RoomTopicStateEventContent,
    SerializableAttrs,
    StateEventContent,
    UserID,
)
from mautrix.util.logging import TraceLogger

from .. import api as as_api, state_store as ss


def quote(*args, **kwargs):
    return urllib_quote(*args, **kwargs, safe="")


_bridgebot = object()

ENSURE_REGISTERED_METHODS = (
    # Room methods
    ClientAPI.create_room,
    ClientAPI.add_room_alias,
    ClientAPI.remove_room_alias,
    ClientAPI.resolve_room_alias,
    ClientAPI.get_joined_rooms,
    StoreUpdatingAPI.join_room_by_id,
    StoreUpdatingAPI.join_room,
    StoreUpdatingAPI.leave_room,
    ClientAPI.set_room_directory_visibility,
    ClientAPI.forget_room,
    # User data methods
    ClientAPI.search_users,
    ClientAPI.set_displayname,
    ClientAPI.set_avatar_url,
    ClientAPI.unstable_create_mxc,
    ClientAPI.upload_media,
    ClientAPI.send_receipt,
    ClientAPI.set_fully_read_marker,
)

ENSURE_JOINED_METHODS = (
    # Room methods
    StoreUpdatingAPI.invite_user,
    # Event methods
    ClientAPI.get_event,
    StoreUpdatingAPI.get_state_event,
    StoreUpdatingAPI.get_state,
    ClientAPI.get_joined_members,
    ClientAPI.get_messages,
    StoreUpdatingAPI.send_state_event,
    ClientAPI.send_message_event,
    ClientAPI.redact,
)

DOUBLE_PUPPET_SOURCE_KEY = "fi.mau.double_puppet_source"

T = TypeVar("T")


class IntentAPI(StoreUpdatingAPI):
    """
    IntentAPI is a high-level wrapper around the AppServiceAPI that provides many easy-to-use
    functions for accessing the client-server API. It is designed for appservices and will
    automatically handle many things like missing invites using the appservice bot.
    """

    api: as_api.AppServiceAPI
    state_store: ss.ASStateStore
    bot: IntentAPI
    log: TraceLogger

    def __init__(
        self,
        mxid: UserID,
        api: as_api.AppServiceAPI,
        bot: IntentAPI = None,
        state_store: ss.ASStateStore = None,
    ) -> None:
        super().__init__(mxid=mxid, api=api, state_store=state_store)
        self.bot = bot
        self.log = api.base_log.getChild("intent")

        for method in ENSURE_REGISTERED_METHODS:
            method = getattr(self, method.__name__)

            async def wrapper(*args, __self=self, __method=method, **kwargs):
                await __self.ensure_registered()
                return await __method(*args, **kwargs)

            setattr(self, method.__name__, wrapper)

        for method in ENSURE_JOINED_METHODS:
            method = getattr(self, method.__name__)

            async def wrapper(*args, __self=self, __method=method, **kwargs):
                room_id = kwargs.get("room_id", None)
                if not room_id:
                    room_id = args[0]
                ensure_joined = kwargs.pop("ensure_joined", True)
                if ensure_joined:
                    await __self.ensure_joined(room_id)
                return await __method(*args, **kwargs)

            setattr(self, method.__name__, wrapper)

    def user(
        self, user_id: UserID, token: str | None = None, base_url: str | None = None
    ) -> IntentAPI:
        """
        Get the intent API for a specific user.
        This is just a proxy to :meth:`AppServiceAPI.intent`.

        You should only call this method for the bot user. Calling it with child intent APIs will
        result in a warning log.

        Args:
            user_id: The Matrix ID of the user whose intent API to get.
            token: The access token to use for the Matrix ID.
            base_url: An optional URL to use for API requests.

        Returns:
            The IntentAPI for the given user.
        """
        if not self.bot:
            return self.api.intent(user_id, token, base_url)
        else:
            self.log.warning("Called IntentAPI#user() of child intent object.")
            return self.bot.api.intent(user_id, token, base_url)

    # region User actions

    async def set_presence(
        self,
        presence: PresenceState = PresenceState.ONLINE,
        status: str | None = None,
        ignore_cache: bool = False,
    ):
        """
        Set the online status of the user.

        See also: `API reference <https://matrix.org/docs/spec/client_server/r0.4.0.html#put-matrix-client-r0-presence-userid-status>`__

        Args:
            presence: The online status of the user.
            status: The status message.
            ignore_cache: Whether or not to set presence even if the cache says the presence is
                already set to that value.
        """
        await self.ensure_registered()
        if not ignore_cache and self.state_store.has_presence(self.mxid, status):
            return
        await super().set_presence(presence, status)
        self.state_store.set_presence(self.mxid, status)

    # endregion
    # region Room actions

    def _add_source_key(self, content: T = None) -> T:
        if self.api.is_real_user and self.api.bridge_name:
            if not content:
                content = {}
            content[DOUBLE_PUPPET_SOURCE_KEY] = self.api.bridge_name
        return content

    async def invite_user(
        self,
        room_id: RoomID,
        user_id: UserID,
        reason: str | None = None,
        check_cache: bool = False,
        extra_content: dict[str, Any] | None = None,
    ) -> None:
        """
        Invite a user to participate in a particular room. They do not start participating in the
        room until they actually join the room.

        Only users currently in the room can invite other users to join that room.

        If the user was invited to the room, the homeserver will add a `m.room.member`_ event to
        the room.

        See also: `API reference <https://spec.matrix.org/v1.1/client-server-api/#post_matrixclientv3roomsroomidinvite>`__

        .. _m.room.member: https://spec.matrix.org/v1.1/client-server-api/#mroommember

        Args:
            room_id: The ID of the room to which to invite the user.
            user_id: The fully qualified user ID of the invitee.
            reason: The reason the user was invited. This will be supplied as the ``reason`` on
                the `m.room.member`_ event.
            check_cache: If ``True``, the function will first check the state store, and not do
                         anything if the state store says the user is already invited or joined.
            extra_content: Additional properties for the invite event content.
                If a non-empty dict is passed, the invite event will be created using
                the ``PUT /state/m.room.member/...`` endpoint instead of ``POST /invite``.
        """
        try:
            ok_states = (Membership.INVITE, Membership.JOIN)
            do_invite = not check_cache or (
                await self.state_store.get_membership(room_id, user_id) not in ok_states
            )
            if do_invite:
                extra_content = self._add_source_key(extra_content)
                await super().invite_user(
                    room_id, user_id, reason=reason, extra_content=extra_content
                )
                await self.state_store.invited(room_id, user_id)
        except MAlreadyJoined as e:
            await self.state_store.joined(room_id, user_id)
        except MatrixRequestError as e:
            # TODO remove this once MSC3848 is released and minimum spec version is bumped
            if e.errcode == "M_FORBIDDEN" and "is already in the room" in e.message:
                await self.state_store.joined(room_id, user_id)
            else:
                raise

    async def kick_user(
        self,
        room_id: RoomID,
        user_id: UserID,
        reason: str = "",
        extra_content: dict[str, JSON] | None = None,
    ) -> None:
        extra_content = self._add_source_key(extra_content)
        await super().kick_user(room_id, user_id, reason=reason, extra_content=extra_content)

    async def ban_user(
        self,
        room_id: RoomID,
        user_id: UserID,
        reason: str = "",
        extra_content: dict[str, JSON] | None = None,
    ) -> None:
        extra_content = self._add_source_key(extra_content)
        await super().ban_user(room_id, user_id, reason=reason, extra_content=extra_content)

    async def unban_user(
        self,
        room_id: RoomID,
        user_id: UserID,
        reason: str = "",
        extra_content: dict[str, JSON] | None = None,
    ) -> None:
        extra_content = self._add_source_key(extra_content)
        await super().unban_user(room_id, user_id, reason=reason, extra_content=extra_content)

    async def join_room_by_id(
        self,
        room_id: RoomID,
        third_party_signed: JSON = None,
        extra_content: dict[str, JSON] | None = None,
    ) -> RoomID:
        extra_content = self._add_source_key(extra_content)
        return await super().join_room_by_id(
            room_id, third_party_signed=third_party_signed, extra_content=extra_content
        )

    async def leave_room(
        self,
        room_id: RoomID,
        reason: str | None = None,
        extra_content: dict[str, JSON] | None = None,
        raise_not_in_room: bool = False,
    ) -> None:
        extra_content = self._add_source_key(extra_content)
        await super().leave_room(room_id, reason, extra_content, raise_not_in_room)

    def set_room_avatar(
        self, room_id: RoomID, avatar_url: ContentURI | None, **kwargs
    ) -> Awaitable[EventID]:
        content = RoomAvatarStateEventContent(url=avatar_url)
        content = self._add_source_key(content)
        return self.send_state_event(room_id, EventType.ROOM_AVATAR, content, **kwargs)

    def set_room_name(self, room_id: RoomID, name: str, **kwargs) -> Awaitable[EventID]:
        content = RoomNameStateEventContent(name=name)
        content = self._add_source_key(content)
        return self.send_state_event(room_id, EventType.ROOM_NAME, content, **kwargs)

    def set_room_topic(self, room_id: RoomID, topic: str, **kwargs) -> Awaitable[EventID]:
        content = RoomTopicStateEventContent(topic=topic)
        content = self._add_source_key(content)
        return self.send_state_event(room_id, EventType.ROOM_TOPIC, content, **kwargs)

    async def get_power_levels(
        self, room_id: RoomID, ignore_cache: bool = False, ensure_joined: bool = True
    ) -> PowerLevelStateEventContent:
        if ensure_joined:
            await self.ensure_joined(room_id)
        if not ignore_cache:
            levels = await self.state_store.get_power_levels(room_id)
            if levels:
                return levels
        try:
            levels = await self.get_state_event(room_id, EventType.ROOM_POWER_LEVELS)
        except MNotFound:
            levels = PowerLevelStateEventContent()
        except MForbidden:
            if not ensure_joined:
                return PowerLevelStateEventContent()
            raise
        await self.state_store.set_power_levels(room_id, levels)
        return levels

    async def set_power_levels(
        self, room_id: RoomID, content: PowerLevelStateEventContent, **kwargs
    ) -> EventID:
        content = self._add_source_key(content)
        response = await self.send_state_event(
            room_id, EventType.ROOM_POWER_LEVELS, content, **kwargs
        )
        if response:
            await self.state_store.set_power_levels(room_id, content)
        return response

    async def get_pinned_messages(self, room_id: RoomID) -> list[EventID]:
        await self.ensure_joined(room_id)
        try:
            content = await self.get_state_event(room_id, EventType.ROOM_PINNED_EVENTS)
        except MNotFound:
            return []
        return content["pinned"]

    def set_pinned_messages(
        self, room_id: RoomID, events: list[EventID], **kwargs
    ) -> Awaitable[EventID]:
        content = RoomPinnedEventsStateEventContent(pinned=events)
        content = self._add_source_key(content)
        return self.send_state_event(room_id, EventType.ROOM_PINNED_EVENTS, content, **kwargs)

    async def pin_message(self, room_id: RoomID, event_id: EventID) -> None:
        events = await self.get_pinned_messages(room_id)
        if event_id not in events:
            events.append(event_id)
            await self.set_pinned_messages(room_id, events)

    async def unpin_message(self, room_id: RoomID, event_id: EventID):
        events = await self.get_pinned_messages(room_id)
        if event_id in events:
            events.remove(event_id)
            await self.set_pinned_messages(room_id, events)

    async def set_join_rule(self, room_id: RoomID, join_rule: JoinRule, **kwargs):
        content = JoinRulesStateEventContent(join_rule=join_rule)
        content = self._add_source_key(content)
        await self.send_state_event(room_id, EventType.ROOM_JOIN_RULES, content, **kwargs)

    async def get_room_displayname(
        self, room_id: RoomID, user_id: UserID, ignore_cache=False
    ) -> str:
        return (await self.get_room_member_info(room_id, user_id, ignore_cache)).displayname

    async def get_room_avatar_url(
        self, room_id: RoomID, user_id: UserID, ignore_cache=False
    ) -> str:
        return (await self.get_room_member_info(room_id, user_id, ignore_cache)).avatar_url

    async def get_room_member_info(
        self, room_id: RoomID, user_id: UserID, ignore_cache=False
    ) -> Member:
        member = await self.state_store.get_member(room_id, user_id)
        if not member or not member.membership or ignore_cache:
            member = await self.get_state_event(room_id, EventType.ROOM_MEMBER, user_id)
        return member

    async def set_typing(
        self,
        room_id: RoomID,
        is_typing: bool = True,
        timeout: int = 5000,
        ignore_cache: bool = False,
    ) -> None:
        await self.ensure_joined(room_id)
        if not ignore_cache and is_typing == self.state_store.is_typing(room_id, self.mxid):
            return
        await super().set_typing(room_id, timeout if is_typing else 0)
        self.state_store.set_typing(room_id, self.mxid, is_typing, timeout)

    async def error_and_leave(
        self, room_id: RoomID, text: str | None = None, html: str | None = None
    ) -> None:
        await self.ensure_joined(room_id)
        await self.send_notice(room_id, text, html=html)
        await self.leave_room(room_id)

    async def send_message_event(
        self, room_id: RoomID, event_type: EventType, content: EventContent, **kwargs
    ) -> EventID:
        await self._ensure_has_power_level_for(room_id, event_type)
        content = self._add_source_key(content)
        return await super().send_message_event(room_id, event_type, content, **kwargs)

    async def redact(
        self,
        room_id: RoomID,
        event_id: EventID,
        reason: str | None = None,
        extra_content: dict[str, JSON] | None = None,
        **kwargs,
    ) -> EventID:
        await self._ensure_has_power_level_for(room_id, EventType.ROOM_REDACTION)
        extra_content = self._add_source_key(extra_content)
        return await super().redact(
            room_id, event_id, reason, extra_content=extra_content, **kwargs
        )

    async def send_state_event(
        self,
        room_id: RoomID,
        event_type: EventType,
        content: StateEventContent | dict[str, JSON],
        state_key: str = "",
        **kwargs,
    ) -> EventID:
        await self._ensure_has_power_level_for(room_id, event_type, state_key=state_key)
        content = self._add_source_key(content)
        return await super().send_state_event(room_id, event_type, content, state_key, **kwargs)

    async def get_room_members(
        self, room_id: RoomID, allowed_memberships: tuple[Membership, ...] = (Membership.JOIN,)
    ) -> list[UserID]:
        if len(allowed_memberships) == 1 and allowed_memberships[0] == Membership.JOIN:
            memberships = await self.get_joined_members(room_id)
            return list(memberships.keys())
        member_events = await self.get_members(room_id)
        return [
            UserID(evt.state_key)
            for evt in member_events
            if evt.content.membership in allowed_memberships
        ]

    async def mark_read(
        self, room_id: RoomID, event_id: EventID, extra_content: dict[str, JSON] | None = None
    ) -> None:
        if self.state_store.get_read(room_id, self.mxid) != event_id:
            if self.api.is_real_user and self.api.bridge_name:
                if not extra_content:
                    extra_content = {}
                double_puppet_indicator = {
                    DOUBLE_PUPPET_SOURCE_KEY: self.api.bridge_name,
                }
                extra_content.update(
                    {
                        "com.beeper.fully_read.extra": double_puppet_indicator,
                        "com.beeper.read.extra": double_puppet_indicator,
                    }
                )
            await self.set_fully_read_marker(
                room_id,
                fully_read=event_id,
                read_receipt=event_id,
                extra_content=extra_content,
            )
            self.state_store.set_read(room_id, self.mxid, event_id)

    async def batch_send(
        self,
        room_id: RoomID,
        prev_event_id: EventID,
        *,
        batch_id: BatchID | None = None,
        events: Iterable[BatchSendEvent],
        state_events_at_start: Iterable[BatchSendStateEvent] = (),
        beeper_new_messages: bool = False,
    ) -> BatchSendResponse:
        """
        Send a batch of historical events into a room. See `MSC2716`_ for more info.

        .. _MSC2716: https://github.com/matrix-org/matrix-doc/pull/2716

        .. versionadded:: v0.12.5

        Args:
            room_id: The room ID to send the events to.
            prev_event_id: The anchor event. The batch will be inserted immediately after this event.
            batch_id: The batch ID for sending a continuation of an earlier batch. If provided,
                      the new batch will be inserted between the prev event and the previous batch.
            events: The events to send.
            state_events_at_start: The state events to send at the start of the batch.
                                   These will be sent as outlier events, which means they won't be
                                   a part of the actual room state.
            beeper_new_messages: Custom flag to tell the server that the messages can be sent to
                                 the end of the room as normal messages instead of history.

        Returns:
            All the event IDs generated, plus a batch ID that can be passed back to this method.
        """
        path = Path.unstable["org.matrix.msc2716"].rooms[room_id].batch_send
        query: JSON = {"prev_event_id": prev_event_id}
        if batch_id:
            query["batch_id"] = batch_id
        if beeper_new_messages:
            query["com.beeper.new_messages"] = "true"
        resp = await self.api.request(
            Method.POST,
            path,
            query_params=query,
            content={
                "events": [evt.serialize() for evt in events],
                "state_events_at_start": [evt.serialize() for evt in state_events_at_start],
            },
        )
        return BatchSendResponse.deserialize(resp)

    # endregion
    # region Ensure functions

    async def ensure_joined(
        self, room_id: RoomID, ignore_cache: bool = False, bot: IntentAPI | None = _bridgebot
    ) -> bool:
        """
        Ensure the user controlled by this intent is joined to the given room.

        If the user is not in the room and the room is invite-only or the user is banned, this will
        first invite and/or unban the user using the bridge bot account.

        Args:
            room_id: The room to join.
            ignore_cache: Should the Matrix state store be checked first?
                If ``False`` and the store says the user is in the room, no requests will be made.
            bot: An optional override account to use as the bridge bot. This is useful if you know
                the bridge bot is not an admin in the room, but some other ghost user is.

        Returns:
            ``False`` if the cache said the user is already in the room,
            ``True`` if the user was successfully added to the room just now.
        """
        if not room_id:
            raise ValueError("Room ID not given")
        if not ignore_cache and await self.state_store.is_joined(room_id, self.mxid):
            return False
        if bot is _bridgebot:
            bot = self.bot
        if bot is self:
            bot = None
        await self.ensure_registered()
        try:
            await self.join_room(room_id, max_retries=0)
            await self.state_store.joined(room_id, self.mxid)
        except MForbidden as e:
            if not bot:
                raise IntentError(f"Failed to join room {room_id} as {self.mxid}") from e
            try:
                await bot.invite_user(room_id, self.mxid)
                await self.join_room(room_id, max_retries=0)
                await self.state_store.joined(room_id, self.mxid)
            except MatrixRequestError as e2:
                raise IntentError(f"Failed to join room {room_id} as {self.mxid}") from e2
        except MBadState as e:
            if not bot:
                raise IntentError(f"Failed to join room {room_id} as {self.mxid}") from e
            try:
                await bot.unban_user(room_id, self.mxid)
                await bot.invite_user(room_id, self.mxid)
                await self.join_room(room_id, max_retries=0)
                await self.state_store.joined(room_id, self.mxid)
            except MatrixRequestError as e2:
                raise IntentError(f"Failed to join room {room_id} as {self.mxid}") from e2
        except MatrixRequestError as e:
            raise IntentError(f"Failed to join room {room_id} as {self.mxid}") from e
        return True

    def _register(self) -> Awaitable[dict]:
        content = {
            "username": self.localpart,
            "type": "m.login.application_service",
            "inhibit_login": True,
        }
        query_params = {"kind": "user"}
        return self.api.request(Method.POST, Path.v3.register, content, query_params=query_params)

    async def ensure_registered(self) -> None:
        """
        Ensure the user controlled by this intent has been registered on the homeserver.

        This will always check the state store first, but the ``M_USER_IN_USE`` error will also be
        silently ignored, so it's fine if the state store isn't accurate. However, if using double
        puppeting, the state store should always return ``True`` for those users.
        """
        if await self.state_store.is_registered(self.mxid):
            return
        try:
            await self._register()
        except MUserInUse:
            pass
        await self.state_store.registered(self.mxid)

    async def _ensure_has_power_level_for(
        self, room_id: RoomID, event_type: EventType, state_key: str = ""
    ) -> None:
        if not room_id:
            raise ValueError("Room ID not given")
        elif not event_type:
            raise ValueError("Event type not given")

        if event_type == EventType.ROOM_MEMBER:
            # TODO: if state_key doesn't equal self.mxid, check invite/kick/ban permissions
            return
        if not await self.state_store.has_power_levels_cached(room_id):
            # TODO add option to not try to fetch power levels from server
            await self.get_power_levels(room_id, ignore_cache=True, ensure_joined=False)
        if not await self.state_store.has_power_level(room_id, self.mxid, event_type):
            # TODO implement something better
            raise IntentError(
                f"Power level of {self.mxid} is not enough for {event_type} in {room_id}"
            )
            # self.log.warning(
            #     f"Power level of {self.mxid} is not enough for {event_type} in {room_id}")

    # endregion



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from typing import Any, AsyncContextManager, Type
from abc import ABC, abstractmethod
import logging

from yarl import URL

from mautrix import __optional_imports__
from mautrix.util.logging import TraceLogger

from .connection import LoggingConnection
from .errors import DatabaseNotOwned, ForeignTablesFound
from .scheme import Scheme
from .upgrade import UpgradeTable, upgrade_tables

if __optional_imports__:
    from aiosqlite import Cursor
    from asyncpg import Record


class Database(ABC):
    schemes: dict[str, Type[Database]] = {}
    log: TraceLogger

    scheme: Scheme
    url: URL
    _db_args: dict[str, Any]
    upgrade_table: UpgradeTable | None
    owner_name: str | None
    ignore_foreign_tables: bool

    def __init__(
        self,
        url: URL,
        upgrade_table: UpgradeTable | None,
        db_args: dict[str, Any] | None = None,
        log: TraceLogger | None = None,
        owner_name: str | None = None,
        ignore_foreign_tables: bool = True,
    ) -> None:
        self.url = url
        self._db_args = {**db_args} if db_args else {}
        self.upgrade_table = upgrade_table
        self.owner_name = owner_name
        self.ignore_foreign_tables = ignore_foreign_tables
        self.log = log or logging.getLogger("mau.db")
        assert isinstance(self.log, TraceLogger)

    @classmethod
    def create(
        cls,
        url: str | URL,
        *,
        db_args: dict[str, Any] | None = None,
        upgrade_table: UpgradeTable | str | None = None,
        log: logging.Logger | TraceLogger | None = None,
        owner_name: str | None = None,
        ignore_foreign_tables: bool = True,
    ) -> Database:
        url = URL(url)
        try:
            impl = cls.schemes[url.scheme]
        except KeyError as e:
            if url.scheme in ("postgres", "postgresql"):
                raise RuntimeError(
                    f"Unknown database scheme {url.scheme}."
                    " Perhaps you forgot to install asyncpg?"
                ) from e
            elif url.scheme in ("sqlite", "sqlite3"):
                raise RuntimeError(
                    f"Unknown database scheme {url.scheme}."
                    " Perhaps you forgot to install aiosqlite?"
                ) from e
            raise RuntimeError(f"Unknown database scheme {url.scheme}") from e
        if isinstance(upgrade_table, str):
            upgrade_table = upgrade_tables[upgrade_table]
        elif upgrade_table is None:
            upgrade_table = UpgradeTable()
        elif not isinstance(upgrade_table, UpgradeTable):
            raise ValueError(f"Can't use {type(upgrade_table)} as the upgrade table")
        return impl(
            url,
            db_args=db_args,
            upgrade_table=upgrade_table,
            log=log,
            owner_name=owner_name,
            ignore_foreign_tables=ignore_foreign_tables,
        )

    def override_pool(self, db: Database) -> None:
        pass

    async def start(self) -> None:
        if not self.ignore_foreign_tables:
            await self._check_foreign_tables()
        if self.owner_name:
            await self._check_owner()
        if self.upgrade_table and len(self.upgrade_table.upgrades) > 0:
            await self.upgrade_table.upgrade(self)

    async def _check_foreign_tables(self) -> None:
        if await self.table_exists("state_groups_state"):
            raise ForeignTablesFound("found state_groups_state likely belonging to Synapse")
        elif await self.table_exists("roomserver_rooms"):
            raise ForeignTablesFound("found roomserver_rooms likely belonging to Dendrite")

    async def _check_owner(self) -> None:
        await self.execute(
            """CREATE TABLE IF NOT EXISTS database_owner (
                key   INTEGER PRIMARY KEY DEFAULT 0,
                owner TEXT NOT NULL
            )"""
        )
        owner = await self.fetchval("SELECT owner FROM database_owner WHERE key=0")
        if not owner:
            await self.execute("INSERT INTO database_owner (owner) VALUES ($1)", self.owner_name)
        elif owner != self.owner_name:
            raise DatabaseNotOwned(owner)

    @abstractmethod
    async def stop(self) -> None:
        pass

    @abstractmethod
    def acquire(self) -> AsyncContextManager[LoggingConnection]:
        pass

    async def execute(self, query: str, *args: Any, timeout: float | None = None) -> str | Cursor:
        async with self.acquire() as conn:
            return await conn.execute(query, *args, timeout=timeout)

    async def executemany(
        self, query: str, *args: Any, timeout: float | None = None
    ) -> str | Cursor:
        async with self.acquire() as conn:
            return await conn.executemany(query, *args, timeout=timeout)

    async def fetch(self, query: str, *args: Any, timeout: float | None = None) -> list[Record]:
        async with self.acquire() as conn:
            return await conn.fetch(query, *args, timeout=timeout)

    async def fetchval(
        self, query: str, *args: Any, column: int = 0, timeout: float | None = None
    ) -> Any:
        async with self.acquire() as conn:
            return await conn.fetchval(query, *args, column=column, timeout=timeout)

    async def fetchrow(
        self, query: str, *args: Any, timeout: float | None = None
    ) -> Record | None:
        async with self.acquire() as conn:
            return await conn.fetchrow(query, *args, timeout=timeout)

    async def table_exists(self, name: str) -> bool:
        async with self.acquire() as conn:
            return await conn.table_exists(name)



from .api import DOUBLE_PUPPET_SOURCE_KEY, AppServiceAPI, ChildAppServiceAPI, IntentAPI
from .appservice import AppService
from .as_handler import AppServiceServerMixin
from .state_store import ASStateStore

__all__ = [
    "AppService",
    "AppServiceAPI",
    "ChildAppServiceAPI",
    "IntentAPI",
    "ASStateStore",
    "AppServiceServerMixin",
    "DOUBLE_PUPPET_SOURCE_KEY",
    "state_store",
]



# This is a script that fixes the __module__ tags in mautrix-python and some libraries.
# It's used to help Sphinx/autodoc figure out where the things are canonically imported from
# (by default, it shows the exact module they're defined in rather than the top-level import path).

from typing import NewType
from types import FunctionType, ModuleType

import aiohttp

from . import appservice, bridge, client, crypto, errors, types
from .crypto import attachments
from .util import async_db, config, db, formatter, logging


def _fix(obj: ModuleType) -> None:
    for item_name in getattr(obj, "__all__", None) or dir(obj):
        item = getattr(obj, item_name)
        if isinstance(item, (type, FunctionType, NewType)):
            # Ignore backwards-compatibility imports like the BridgeState import in mautrix.bridge
            if item.__module__.startswith("mautrix") and not item.__module__.startswith(
                obj.__name__
            ):
                continue

            item.__module__ = obj.__name__

            if isinstance(item, NewType):
                # By default autodoc makes a blank "Bases:" text,
                # so adjust it to show the type as the "base"
                item.__bases__ = (item.__supertype__,)
        # elif type(item).__module__ == "typing":
        #     print(obj.__name__, item_name, type(item))


_things_to_fix = [
    types,
    bridge,
    client,
    crypto,
    attachments,
    errors,
    appservice,
    async_db,
    config,
    db,
    formatter,
    logging,
    aiohttp,
]

for mod in _things_to_fix:
    _fix(mod)



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from .entity_string import AbstractEntity, EntityString, SemiAbstractEntity, SimpleEntity
from .formatted_string import EntityType, FormattedString
from .html_reader import HTMLNode, read_html
from .markdown_string import MarkdownString
from .parser import MatrixParser, RecursionContext


async def parse_html(input_html: str) -> str:
    return (await MatrixParser().parse(input_html)).text


__all__ = [
    "AbstractEntity",
    "EntityString",
    "SemiAbstractEntity",
    "SimpleEntity",
    "EntityType",
    "FormattedString",
    "HTMLNode",
    "read_html",
    "MarkdownString",
    "MatrixParser",
    "RecursionContext",
    "parse_html",
]



from mautrix import __optional_imports__

from .abstract import CryptoStore, StateStore
from .memory import MemoryCryptoStore

try:
    from .asyncpg import PgCryptoStateStore, PgCryptoStore
except ImportError:
    if __optional_imports__:
        raise
    PgCryptoStore = PgCryptoStateStore = None

__all__ = ["CryptoStore", "StateStore", "MemoryCryptoStore", "PgCryptoStateStore", "PgCryptoStore"]



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from mautrix.errors import IntentError, MatrixRequestError, MForbidden
from mautrix.types import ContentURI, EventID, UserID

from ... import bridge as br
from .handler import SECTION_ADMIN, CommandEvent, command_handler


@command_handler(
    needs_admin=True,
    needs_auth=False,
    name="set-pl",
    help_section=SECTION_ADMIN,
    help_args="[_mxid_] <_level_>",
    help_text="Set a temporary power level without affecting the remote platform.",
)
async def set_power_level(evt: CommandEvent) -> EventID:
    try:
        user_id = UserID(evt.args[0])
    except IndexError:
        return await evt.reply(f"**Usage:** `$cmdprefix+sp set-pl [mxid] <level>`")

    if user_id.startswith("@"):
        evt.args.pop(0)
    else:
        user_id = evt.sender.mxid

    try:
        level = int(evt.args[0])
    except (KeyError, IndexError):
        return await evt.reply("**Usage:** `$cmdprefix+sp set-pl [mxid] <level>`")
    except ValueError:
        return await evt.reply("The level must be an integer.")
    levels = await evt.main_intent.get_power_levels(evt.room_id, ignore_cache=True)
    levels.users[user_id] = level
    try:
        return await evt.main_intent.set_power_levels(evt.room_id, levels)
    except MForbidden as e:
        await evt.reply(f"I don't seem to have permission to update power levels: {e.message}")
    except (MatrixRequestError, IntentError):
        evt.log.exception("Failed to update power levels")
        return await evt.reply("Failed to update power levels (see logs for more details)")


async def _get_mxid_param(
    evt: CommandEvent, args: str
) -> tuple[br.BasePuppet | None, EventID | None]:
    try:
        user_id = UserID(evt.args[0])
    except IndexError:
        return None, await evt.reply(f"**Usage:** `$cmdprefix+sp {evt.command} {args}`")

    if user_id.startswith("@") and ":" in user_id:
        # TODO support parsing mention pills instead of requiring a plaintext mxid
        puppet = await evt.bridge.get_puppet(user_id)
        if not puppet:
            return None, await evt.reply("The given user ID is not a valid ghost user.")
        evt.args.pop(0)
        return puppet, None
    elif evt.is_portal and (puppet := await evt.portal.get_dm_puppet()):
        return puppet, None
    return None, await evt.reply(
        "This is not a private chat portal, you must pass a user ID explicitly."
    )


@command_handler(
    needs_admin=True,
    needs_auth=False,
    name="set-avatar",
    help_section=SECTION_ADMIN,
    help_args="[_mxid_] <_mxc:// uri_>",
    help_text="Set an avatar for a ghost user.",
)
async def set_ghost_avatar(evt: CommandEvent) -> EventID | None:
    puppet, err = await _get_mxid_param(evt, "[mxid] <mxc:// URI>")
    if err:
        return err

    try:
        mxc_uri = ContentURI(evt.args[0])
    except IndexError:
        return await evt.reply("**Usage:** `$cmdprefix+sp set-avatar [mxid] <mxc_uri>`")
    if not mxc_uri.startswith("mxc://"):
        return await evt.reply("The avatar URL must start with `mxc://`")

    try:
        return await puppet.default_mxid_intent.set_avatar_url(mxc_uri)
    except (MatrixRequestError, IntentError):
        evt.log.exception("Failed to set avatar.")
        return await evt.reply("Failed to set avatar (see logs for more details).")


@command_handler(
    needs_admin=True,
    needs_auth=False,
    name="remove-avatar",
    help_section=SECTION_ADMIN,
    help_args="[_mxid_]",
    help_text="Remove the avatar for a ghost user.",
)
async def remove_ghost_avatar(evt: CommandEvent) -> EventID | None:
    puppet, err = await _get_mxid_param(evt, "[mxid]")
    if err:
        return err
    try:
        return await puppet.default_mxid_intent.set_avatar_url(ContentURI(""))
    except (MatrixRequestError, IntentError):
        evt.log.exception("Failed to remove avatar.")
        return await evt.reply("Failed to remove avatar (see logs for more details).")


@command_handler(
    needs_admin=True,
    needs_auth=False,
    name="set-displayname",
    help_section=SECTION_ADMIN,
    help_args="[_mxid_] <_displayname_>",
    help_text="Set the display name for a ghost user.",
)
async def set_ghost_display_name(evt: CommandEvent) -> EventID | None:
    puppet, err = await _get_mxid_param(evt, "[mxid] <displayname>")
    if err:
        return err
    try:
        return await puppet.default_mxid_intent.set_displayname(" ".join(evt.args))
    except (MatrixRequestError, IntentError):
        evt.log.exception("Failed to set display name.")
        return await evt.reply("Failed to set display name (see logs for more details).")


@command_handler(
    needs_admin=True,
    needs_auth=False,
    name="remove-displayname",
    help_section=SECTION_ADMIN,
    help_args="[_mxid_]",
    help_text="Remove the display name for a ghost user.",
)
async def remove_ghost_display_name(evt: CommandEvent) -> EventID | None:
    puppet, err = await _get_mxid_param(evt, "[mxid]")
    if err:
        return err
    try:
        return await puppet.default_mxid_intent.set_displayname("")
    except (MatrixRequestError, IntentError):
        evt.log.exception("Failed to remove display name.")
        return await evt.reply("Failed to remove display name (see logs for more details).")



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Generic, List, Optional, TypeVar, Union

from attr import dataclass
import attr

from ..primitive import JSON, UserID
from ..util import ExtensibleEnum, SerializableAttrs
from .base import BaseRoomEvent
from .type import EventType


class CallDataType(ExtensibleEnum):
    OFFER = "offer"
    ANSWER = "answer"


class CallHangupReason(ExtensibleEnum):
    ICE_FAILED = "ice_failed"
    INVITE_TIMEOUT = "invite_timeout"
    USER_HANGUP = "user_hangup"
    USER_MEDIA_FAILED = "user_media_failed"
    UNKNOWN_ERROR = "unknown_error"


@dataclass
class CallData(SerializableAttrs):
    sdp: str
    type: CallDataType


@dataclass
class CallCandidate(SerializableAttrs):
    candidate: str
    sdp_m_line_index: int = attr.ib(metadata={"json": "sdpMLineIndex"}, default=None)
    sdp_mid: str = attr.ib(metadata={"json": "sdpMid"}, default=None)


@dataclass
class CallInviteEventContent(SerializableAttrs):
    call_id: str
    lifetime: int
    version: int
    offer: CallData
    party_id: Optional[str] = None
    invitee: Optional[UserID] = None


@dataclass
class CallCandidatesEventContent(SerializableAttrs):
    call_id: str
    version: int
    candidates: List[CallCandidate]
    party_id: Optional[str] = None


@dataclass
class CallSelectAnswerEventContent(SerializableAttrs):
    call_id: str
    version: int
    party_id: str
    selected_party_id: str


@dataclass
class CallAnswerEventContent(SerializableAttrs):
    call_id: str
    version: int
    answer: CallData
    party_id: Optional[str] = None


@dataclass
class CallHangupEventContent(SerializableAttrs):
    call_id: str
    version: int
    reason: CallHangupReason = CallHangupReason.USER_HANGUP
    party_id: Optional[str] = None


@dataclass
class CallRejectEventContent(SerializableAttrs):
    call_id: str
    version: int
    party_id: str


@dataclass
class CallNegotiateEventContent(SerializableAttrs):
    call_id: str
    version: int
    lifetime: int
    party_id: str
    description: CallData


type_to_class = {
    EventType.CALL_INVITE: CallInviteEventContent,
    EventType.CALL_CANDIDATES: CallCandidatesEventContent,
    EventType.CALL_SELECT_ANSWER: CallSelectAnswerEventContent,
    EventType.CALL_ANSWER: CallAnswerEventContent,
    EventType.CALL_HANGUP: CallHangupEventContent,
    EventType.CALL_NEGOTIATE: CallNegotiateEventContent,
    EventType.CALL_REJECT: CallRejectEventContent,
}

CallEventContent = Union[
    CallInviteEventContent,
    CallCandidatesEventContent,
    CallAnswerEventContent,
    CallSelectAnswerEventContent,
    CallHangupEventContent,
    CallNegotiateEventContent,
    CallRejectEventContent,
]

T = TypeVar("T", bound=CallEventContent)


@dataclass
class CallEvent(BaseRoomEvent, SerializableAttrs, Generic[T]):
    content: T

    @classmethod
    def deserialize(cls, data: JSON, event_type: Optional[EventType] = None) -> "CallEvent":
        event_type = event_type or EventType.find(data.get("type"))
        data["content"] = type_to_class[event_type].deserialize(data["content"])
        return super().deserialize(data)



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

import functools

import magic

try:
    _from_buffer = functools.partial(magic.from_buffer, mime=True)
    _from_filename = functools.partial(magic.from_file, mime=True)
except AttributeError:
    _from_buffer = lambda data: magic.detect_from_content(data).mime_type
    _from_filename = lambda file: magic.detect_from_filename(file).mime_type


def mimetype(data: bytes | bytearray | str) -> str:
    """
    Uses magic to determine the mimetype of a file on disk or in memory.

    Supports both libmagic's Python bindings and the python-magic package.

    Args:
        data: The file data, either as in-memory bytes or a path to the file as a string.

    Returns:
        The mime type as a string.
    """
    if isinstance(data, str):
        return _from_filename(data)
    elif isinstance(data, bytes):
        return _from_buffer(data)
    elif isinstance(data, bytearray):
        # Magic doesn't like bytearrays directly, so just copy the first 1024 bytes for it.
        return _from_buffer(bytes(data[:1024]))
    else:
        raise TypeError(
            f"mimetype() argument must be a string or bytes, not {type(data).__name__!r}"
        )


__all__ = ["mimetype"]



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from typing import ClassVar, NamedTuple
from enum import Enum
import urllib.parse

from yarl import URL

from .primitive import EventID, RoomAlias, RoomID, UserID
from .util import ExtensibleEnum


class IdentifierType(Enum):
    """The type qualifier for entities in a Matrix URI."""

    EVENT = "$"
    USER = "@"
    ROOM_ALIAS = "#"
    ROOM_ID = "!"

    @property
    def sigil(self) -> str:
        """The sigil of the identifier, used in Matrix events, matrix.to URLs and other places"""
        return _type_to_sigil[self]

    @property
    def uri_type_qualifier(self) -> str:
        """The type qualifier of the identifier, only used in ``matrix:`` URIs."""
        return _type_to_path[self]

    @classmethod
    def from_sigil(cls, sigil: str) -> IdentifierType:
        """Get the IdentifierType corresponding to the given sigil."""
        return _sigil_to_type[sigil]

    @classmethod
    def from_uri_type_qualifier(cls, uri_type_qualifier: str) -> IdentifierType:
        """Get the IdentifierType corresponding to the given ``matrix:`` URI type qualifier."""
        return _path_to_type[uri_type_qualifier]

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}.{self.name}"


class URIAction(ExtensibleEnum):
    """Represents an intent for what the client should do with a Matrix URI."""

    JOIN = "join"
    CHAT = "chat"


_type_to_path: dict[IdentifierType, str] = {
    IdentifierType.EVENT: "e",
    IdentifierType.USER: "u",
    IdentifierType.ROOM_ALIAS: "r",
    IdentifierType.ROOM_ID: "roomid",
}
_path_to_type: dict[str, IdentifierType] = {v: k for k, v in _type_to_path.items()}
_type_to_sigil: dict[IdentifierType, str] = {it: it.value for it in IdentifierType}
_sigil_to_type: dict[str, IdentifierType] = {v: k for k, v in _type_to_sigil.items()}


class _PathPart(NamedTuple):
    type: IdentifierType
    identifier: str

    @classmethod
    def from_mxid(cls, mxid: UserID | RoomID | EventID | RoomAlias | str) -> _PathPart:
        return _PathPart(type=IdentifierType.from_sigil(mxid[0]), identifier=mxid[1:])

    @property
    def mxid(self) -> str:
        return f"{self.type.sigil}{self.identifier}"

    def __str__(self) -> str:
        return self.mxid

    def __repr__(self) -> str:
        return f"_PathPart({self.type!r}, {self.identifier!r})"

    def __eq__(self, other: _PathPart) -> bool:
        if not isinstance(other, _PathPart):
            return False
        return other.type == self.type and other.identifier == self.identifier


_uri_base = URL.build(scheme="matrix")


class MatrixURIError(ValueError):
    """Raised by :meth:`MatrixURI.parse` when parsing a URI fails."""


class MatrixURI:
    """
    A container for Matrix URI data. Supports parsing and generating both ``matrix:`` URIs
    and ``https://matrix.to`` URLs with the same interface.
    """

    URI_BY_DEFAULT: ClassVar[bool] = False
    """Whether :meth:`__str__` should return the matrix: URI instead of matrix.to URL."""

    _part1: _PathPart
    _part2: _PathPart | None

    via: list[str] | None
    """Servers that know about the resource. Important for room ID links."""

    action: URIAction | None
    """The intent for what clients should do with the URI."""

    def __init__(self) -> None:
        """Internal initializer for MatrixURI, external users should use
        either :meth:`build` or :meth:`parse`."""
        self._part2 = None
        self.via = None
        self.action = None

    @classmethod
    def build(
        cls,
        part1: RoomID | UserID | RoomAlias,
        part2: EventID | None = None,
        via: list[str] | None = None,
        action: URIAction | None = None,
    ) -> MatrixURI:
        """
        Construct a MatrixURI instance using an identifier.

        Args:
            part1: The first part of the URI, a user ID, room ID, or room alias.
            part2: The second part of the URI. Only event IDs are allowed,
                   and only allowed when the first part is a room ID or alias.
            via: Servers that know about the resource. Important for room ID links.
            action: The intent for what clients should do with the URI.

        Returns:
            The constructed MatrixURI.

        Raises:
            ValueError: if one of the identifiers doesn't have a valid sigil.

        Examples:
            >>> from mautrix.types import MatrixURI, UserID, RoomAlias, EventID
            >>> MatrixURI.build(UserID("@user:example.com")).matrix_to_url
            'https://matrix.to/#/%40user%3Aexample.com'
            >>> MatrixURI.build(UserID("@user:example.com")).matrix_uri
            'matrix:u/user:example.com'
            >>> # Picks the format based on the URI_BY_DEFAULT field.
            >>> # The default value will be changed to True in a later release.
            >>> str(MatrixURI.build(UserID("@user:example.com")))
            'https://matrix.to/#/%40user%3Aexample.com'
            >>> MatrixURI.build(RoomAlias("#room:example.com"), EventID("$abc123")).matrix_uri
            'matrix:r/room:example.com/e/abc123'
        """
        uri = cls()
        try:
            uri._part1 = _PathPart.from_mxid(part1)
        except KeyError as e:
            raise ValueError(f"Invalid sigil in part 1 '{part1[0]}'") from e
        if uri._part1.type == IdentifierType.EVENT:
            raise ValueError(f"Event ID links must have a room ID or alias too")
        if part2:
            try:
                uri._part2 = _PathPart.from_mxid(part2)
            except KeyError as e:
                raise ValueError(f"Invalid sigil in part 2 '{part2[0]}'") from e
            if uri._part2.type != IdentifierType.EVENT:
                raise ValueError("The second part of Matrix URIs can only be an event ID")
            if uri._part1.type not in (IdentifierType.ROOM_ID, IdentifierType.ROOM_ALIAS):
                raise ValueError("Can't create an event ID link without a room link as the base")
        uri.via = via
        uri.action = action
        return uri

    @classmethod
    def try_parse(cls, url: str | URL) -> MatrixURI | None:
        """
        Try to parse a ``matrix:`` URI or ``https://matrix.to`` URL into parts.
        If parsing fails, return ``None`` instead of throwing an error.

        Args:
            url: The URI to parse, either as a string or a :class:`yarl.URL` instance.

        Returns:
            The parsed data, or ``None`` if parsing failed.
        """
        try:
            return cls.parse(url)
        except ValueError:
            return None

    @classmethod
    def parse(cls, url: str | URL) -> MatrixURI:
        """
        Parse a ``matrix:`` URI or ``https://matrix.to`` URL into parts.

        Args:
            url: The URI to parse, either as a string or a :class:`yarl.URL` instance.

        Returns:
            The parsed data.

        Raises:
            ValueError: if yarl fails to parse the given URL string.
            MatrixURIError: if the URL isn't valid in the Matrix spec.

        Examples:
            >>> from mautrix.types import MatrixURI
            >>> MatrixURI.parse("https://matrix.to/#/@user:example.com").user_id
            '@user:example.com'
            >>> MatrixURI.parse("https://matrix.to/#/#room:example.com/$abc123").event_id
            '$abc123'
            >>> MatrixURI.parse("matrix:r/room:example.com/e/abc123").event_id
            '$abc123'
        """
        url = URL(url)
        if url.scheme == "matrix":
            return cls._parse_matrix_uri(url)
        elif url.scheme == "https" and url.host == "matrix.to":
            return cls._parse_matrix_to_url(url)
        else:
            raise MatrixURIError("Invalid URI (not matrix: nor https://matrix.to)")

    @classmethod
    def _parse_matrix_to_url(cls, url: URL) -> MatrixURI:
        path, *rest = url.raw_fragment.split("?", 1)
        path_parts = path.split("/")
        if len(path_parts) < 2:
            raise MatrixURIError("matrix.to URL doesn't have enough parts")
        # The first component is the blank part between the # and /
        if path_parts[0] != "":
            raise MatrixURIError("first component of matrix.to URL is not empty as expected")
        query = urllib.parse.parse_qs(rest[0] if len(rest) > 0 else "")
        uri = cls()
        part1 = urllib.parse.unquote(path_parts[1])
        if len(part1) < 2:
            raise MatrixURIError(f"Invalid first entity '{part1}' in matrix.to URL")
        try:
            uri._part1 = _PathPart.from_mxid(part1)
        except KeyError as e:
            raise MatrixURIError(
                f"Invalid sigil '{part1[0]}' in first entity of matrix.to URL"
            ) from e
        if len(path_parts) > 2 and len(path_parts[2]) > 0:
            part2 = urllib.parse.unquote(path_parts[2])
            if len(part2) < 2:
                raise MatrixURIError(f"Invalid second entity '{part2}' in matrix.to URL")
            try:
                uri._part2 = _PathPart.from_mxid(part2)
            except KeyError as e:
                raise MatrixURIError(
                    f"Invalid sigil '{part2[0]}' in second entity of matrix.to URL"
                ) from e
        uri.via = query.get("via", None)
        try:
            uri.action = URIAction(query["action"])
        except KeyError:
            pass
        return uri

    @classmethod
    def _parse_matrix_uri(cls, url: URL) -> MatrixURI:
        components = url.raw_path.split("/")
        if len(components) < 2:
            raise MatrixURIError("URI doesn't contain enough parts")
        try:
            type1 = IdentifierType.from_uri_type_qualifier(components[0])
        except KeyError as e:
            raise MatrixURIError(
                f"Invalid type qualifier '{components[0]}' in first entity of matrix: URI"
            ) from e
        if not components[1]:
            raise MatrixURIError("Identifier in first entity of matrix: URI is empty")
        uri = cls()
        uri._part1 = _PathPart(type1, components[1])
        if len(components) >= 3 and components[2]:
            try:
                type2 = IdentifierType.from_uri_type_qualifier(components[2])
            except KeyError as e:
                raise MatrixURIError(
                    f"Invalid type qualifier '{components[2]}' in second entity of matrix: URI"
                ) from e
            if len(components) < 4 or not components[3]:
                raise MatrixURIError("Identifier in second entity of matrix: URI is empty")
            uri._part2 = _PathPart(type2, components[3])
        uri.via = url.query.getall("via", None)
        try:
            uri.action = URIAction(url.query["action"])
        except KeyError:
            pass
        return uri

    @property
    def user_id(self) -> UserID | None:
        """
        Get the user ID from this parsed URI.

        Returns:
            The user ID in this URI, or ``None`` if this is not a link to a user.
        """
        if self._part1.type == IdentifierType.USER:
            return UserID(self._part1.mxid)
        return None

    @property
    def room_id(self) -> RoomID | None:
        """
        Get the room ID from this parsed URI.

        Returns:
            The room ID in this URI, or ``None`` if this is not a link to a room (or event).
        """
        if self._part1.type == IdentifierType.ROOM_ID:
            return RoomID(self._part1.mxid)
        return None

    @property
    def room_alias(self) -> RoomAlias | None:
        """
        Get the room alias from this parsed URI.

        Returns:
            The room alias in this URI, or ``None`` if this is not a link to a room (or event).
        """
        if self._part1.type == IdentifierType.ROOM_ALIAS:
            return RoomAlias(self._part1.mxid)
        return None

    @property
    def event_id(self) -> EventID | None:
        """
        Get the event ID from this parsed URI.

        Returns:
            The event ID in this URI, or ``None`` if this is not a link to an event in a room.
        """
        if (
            self._part2
            and (self.room_id or self.room_alias)
            and self._part2.type == IdentifierType.EVENT
        ):
            return EventID(self._part2.mxid)
        return None

    @property
    def matrix_to_url(self) -> str:
        """
        Convert this parsed URI into a ``https://matrix.to`` URL.

        Returns:
            The link as a matrix.to URL.
        """
        url = f"https://matrix.to/#/{urllib.parse.quote(self._part1.mxid)}"
        if self._part2:
            url += f"/{urllib.parse.quote(self._part2.mxid)}"
        qp = []
        if self.via:
            qp += (("via", server) for server in self.via)
        if self.action:
            qp.append(("action", self.action))
        if qp:
            url += f"?{urllib.parse.urlencode(qp)}"
        return url

    @property
    def matrix_uri(self) -> str:
        """
        Convert this parsed URI into a ``matrix:`` URI.

        Returns:
            The link as a ``matrix:`` URI.
        """
        u = _uri_base / self._part1.type.uri_type_qualifier / self._part1.identifier
        if self._part2:
            u = u / self._part2.type.uri_type_qualifier / self._part2.identifier
        if self.via:
            u = u.update_query({"via": self.via})
        if self.action:
            u = u.update_query({"action": self.action.value})
        return str(u)

    def __str__(self) -> str:
        if self.URI_BY_DEFAULT:
            return self.matrix_uri
        else:
            return self.matrix_to_url

    def __repr__(self) -> str:
        parts = ", ".join(f"{part!r}" for part in (self._part1, self._part2) if part)
        return f"MatrixURI({parts}, via={self.via!r}, action={self.action!r})"

    def __eq__(self, other: MatrixURI) -> bool:
        """
        Checks equality between two parsed Matrix URIs. The order of the via parameters is ignored,
        but otherwise everything has to match exactly.
        """
        if not isinstance(other, MatrixURI):
            return False
        return (
            self._part1 == other._part1
            and self._part2 == other._part2
            and set(self.via or []) == set(other.via or [])
            and self.action == other.action
        )



# Copyright (c) 2022 Tulir Asokan, Sumner Evans
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Any

from attr import dataclass
import attr

from ..primitive import UserID
from ..util import SerializableAttrs
from .base import BaseEvent


@dataclass
class BatchSendEvent(BaseEvent, SerializableAttrs):
    """Base event class for events sent via a batch send request."""

    sender: UserID
    timestamp: int = attr.ib(metadata={"json": "origin_server_ts"})
    content: Any


@dataclass
class BatchSendStateEvent(BatchSendEvent, SerializableAttrs):
    """
    State events to be used as initial state events on batch send events. These never need to be
    deserialized.
    """

    state_key: str



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
class HTMLNode(list[HTMLNode]):
    tag: str
    text: str
    tail: str
    attrib: dict[str, str]
    def __init__(self, tag: str, attrs: list[tuple[str, str]]) -> None: ...

def read_html(data: str) -> HTMLNode: ...



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from typing import Any, NamedTuple
from abc import ABC, abstractmethod
from collections import defaultdict, deque
import asyncio
import logging
import time

from mautrix.api import Method, Path
from mautrix.appservice import AppService
from mautrix.errors import MNotFound
from mautrix.types import EventID, EventType, Membership, MessageType, RoomID, UserID
from mautrix.util.bridge_state import BridgeState, BridgeStateEvent
from mautrix.util.logging import TraceLogger
from mautrix.util.message_send_checkpoint import (
    MessageSendCheckpoint,
    MessageSendCheckpointReportedBy,
    MessageSendCheckpointStatus,
    MessageSendCheckpointStep,
)
from mautrix.util.opt_prometheus import Gauge

from .. import bridge as br

AsmuxPath = Path.unstable["com.beeper.asmux"]


class WrappedTask(NamedTuple):
    task: asyncio.Task | None


class BaseUser(ABC):
    log: TraceLogger = logging.getLogger("mau.user")
    _async_get_locks: dict[Any, asyncio.Lock] = defaultdict(lambda: asyncio.Lock())
    az: AppService
    bridge: br.Bridge
    loop: asyncio.AbstractEventLoop

    is_whitelisted: bool
    is_admin: bool
    relay_whitelisted: bool
    mxid: UserID

    dm_update_lock: asyncio.Lock
    command_status: dict[str, Any] | None
    _metric_value: dict[Gauge, bool]
    _prev_bridge_status: BridgeState | None
    _bridge_state_queue: deque[BridgeState]
    _bridge_state_loop: asyncio.Task | None

    def __init__(self) -> None:
        self.dm_update_lock = asyncio.Lock()
        self.command_status = None
        self._metric_value = defaultdict(lambda: False)
        self._prev_bridge_status = None
        self.log = self.log.getChild(self.mxid)
        self.relay_whitelisted = False
        self._bridge_state_queue = deque()
        self._bridge_state_loop = None

    @abstractmethod
    async def is_logged_in(self) -> bool:
        raise NotImplementedError()

    @abstractmethod
    async def get_puppet(self) -> br.BasePuppet | None:
        """
        Get the ghost that represents this Matrix user on the remote network.

        Returns:
            The puppet entity, or ``None`` if the user is not logged in,
            or it's otherwise not possible to find the remote ghost.
        """
        raise NotImplementedError()

    @abstractmethod
    async def get_portal_with(
        self, puppet: br.BasePuppet, create: bool = True
    ) -> br.BasePortal | None:
        """
        Get a private chat portal between this user and the given ghost.

        Args:
            puppet: The ghost who the portal should be with.
            create: ``True`` if the portal entity should be created if it doesn't exist.

        Returns:
            The portal entity, or ``None`` if it can't be found,
            or doesn't exist and ``create`` is ``False``.
        """

    async def needs_relay(self, portal: br.BasePortal) -> bool:
        return not await self.is_logged_in()

    async def is_in_portal(self, portal: br.BasePortal) -> bool:
        try:
            member_event = await portal.main_intent.get_state_event(
                portal.mxid, EventType.ROOM_MEMBER, self.mxid
            )
        except MNotFound:
            return False
        return member_event and member_event.membership in (Membership.JOIN, Membership.INVITE)

    async def get_direct_chats(self) -> dict[UserID, list[RoomID]]:
        raise NotImplementedError()

    async def update_direct_chats(self, dms: dict[UserID, list[RoomID]] | None = None) -> None:
        """
        Update the m.direct account data of the user.

        Args:
            dms: DMs to _add_ to the list. If not provided, the list is _replaced_ with the result
                of :meth:`get_direct_chats`.
        """
        if not self.bridge.config["bridge.sync_direct_chat_list"]:
            return

        puppet = await self.bridge.get_double_puppet(self.mxid)
        if not puppet or not puppet.is_real_user:
            return

        self.log.debug("Updating m.direct list on homeserver")
        replace = dms is None
        dms = dms or await self.get_direct_chats()
        if self.bridge.homeserver_software.is_asmux:
            # This uses a secret endpoint for atomically updating the DM list
            await puppet.intent.api.request(
                Method.PUT if replace else Method.PATCH,
                AsmuxPath.dms,
                content=dms,
                headers={"X-Asmux-Auth": self.az.as_token},
            )
        else:
            async with self.dm_update_lock:
                try:
                    current_dms = await puppet.intent.get_account_data(EventType.DIRECT)
                except MNotFound:
                    current_dms = {}
                if replace:
                    # Filter away all existing DM statuses with bridge users
                    filtered_dms = {
                        user: rooms
                        for user, rooms in current_dms.items()
                        if not self.bridge.is_bridge_ghost(user)
                    }
                else:
                    filtered_dms = current_dms
                # Add DM statuses for all rooms in our database
                new_dms = {**filtered_dms, **dms}
                if current_dms != new_dms:
                    await puppet.intent.set_account_data(EventType.DIRECT, new_dms)

    def _track_metric(self, metric: Gauge, value: bool) -> None:
        if self._metric_value[metric] != value:
            if value:
                metric.inc(1)
            else:
                metric.dec(1)
            self._metric_value[metric] = value

    async def fill_bridge_state(self, state: BridgeState) -> None:
        state.user_id = self.mxid
        state.fill()

    async def get_bridge_states(self) -> list[BridgeState]:
        raise NotImplementedError()

    async def push_bridge_state(
        self,
        state_event: BridgeStateEvent,
        error: str | None = None,
        message: str | None = None,
        ttl: int | None = None,
        remote_id: str | None = None,
        info: dict[str, Any] | None = None,
        reason: str | None = None,
    ) -> None:
        if not self.bridge.config["homeserver.status_endpoint"]:
            return

        state = BridgeState(
            state_event=state_event,
            error=error,
            message=message,
            ttl=ttl,
            remote_id=remote_id,
            info=info,
            reason=reason,
        )
        await self.fill_bridge_state(state)
        if state.should_deduplicate(self._prev_bridge_status):
            return
        self._prev_bridge_status = state
        self._bridge_state_queue.append(state)
        if not self._bridge_state_loop or self._bridge_state_loop.done():
            self.log.trace(f"Starting bridge state loop")
            self._bridge_state_loop = asyncio.create_task(self._start_bridge_state_send_loop())
        else:
            self.log.debug(f"Queued bridge state to send later: {state.state_event}")

    async def _start_bridge_state_send_loop(self):
        url = self.bridge.config["homeserver.status_endpoint"]
        while self._bridge_state_queue:
            state = self._bridge_state_queue.popleft()
            success = await state.send(url, self.az.as_token, self.log)
            if not success:
                if state.send_attempts_ <= 10:
                    retry_seconds = state.send_attempts_**2
                    self.log.warning(
                        f"Attempt #{state.send_attempts_} of sending bridge state "
                        f"{state.state_event} failed, retrying in {retry_seconds} seconds"
                    )
                    await asyncio.sleep(retry_seconds)
                    self._bridge_state_queue.appendleft(state)
                else:
                    self.log.error(
                        f"Failed to send bridge state {state.state_event} "
                        f"after {state.send_attempts_} attempts, giving up"
                    )
        self._bridge_state_loop = None

    def send_remote_checkpoint(
        self,
        status: MessageSendCheckpointStatus,
        event_id: EventID,
        room_id: RoomID,
        event_type: EventType,
        message_type: MessageType | None = None,
        error: str | Exception | None = None,
        retry_num: int = 0,
    ) -> WrappedTask:
        """
        Send a remote checkpoint for the given ``event_id``. This function spaws an
        :class:`asyncio.Task`` to send the checkpoint.

        :returns: the checkpoint send task. This can be awaited if you want to block on the
        checkpoint send.
        """
        if not self.bridge.config["homeserver.message_send_checkpoint_endpoint"]:
            return WrappedTask(task=None)
        task = asyncio.create_task(
            MessageSendCheckpoint(
                event_id=event_id,
                room_id=room_id,
                step=MessageSendCheckpointStep.REMOTE,
                timestamp=int(time.time() * 1000),
                status=status,
                reported_by=MessageSendCheckpointReportedBy.BRIDGE,
                event_type=event_type,
                message_type=message_type,
                info=str(error) if error else None,
                retry_num=retry_num,
            ).send(
                self.bridge.config["homeserver.message_send_checkpoint_endpoint"],
                self.az.as_token,
                self.log,
            )
        )
        return WrappedTask(task=task)



from .base import IntentError, MatrixConnectionError, MatrixError, MatrixResponseError
from .crypto import (
    CryptoError,
    DecryptedPayloadError,
    DecryptionError,
    DeviceValidationError,
    DuplicateMessageIndex,
    EncryptionError,
    MatchingSessionDecryptionError,
    MismatchingRoomError,
    SessionNotFound,
    SessionShareError,
    VerificationError,
)
from .request import (
    MAlreadyJoined,
    MatrixBadContent,
    MatrixBadRequest,
    MatrixInvalidToken,
    MatrixRequestError,
    MatrixStandardRequestError,
    MatrixUnknownRequestError,
    MBadJSON,
    MBadState,
    MCaptchaInvalid,
    MCaptchaNeeded,
    MExclusive,
    MForbidden,
    MGuestAccessForbidden,
    MIncompatibleRoomVersion,
    MInsufficientPower,
    MInvalidParam,
    MInvalidRoomState,
    MInvalidUsername,
    MLimitExceeded,
    MMissingParam,
    MMissingToken,
    MNotFound,
    MNotJoined,
    MNotJSON,
    MRoomInUse,
    MTooLarge,
    MUnauthorized,
    MUnknown,
    MUnknownEndpoint,
    MUnknownToken,
    MUnrecognized,
    MUnsupportedRoomVersion,
    MUserDeactivated,
    MUserInUse,
    make_request_error,
    standard_error,
)
from .well_known import (
    WellKnownError,
    WellKnownInvalidVersionsResponse,
    WellKnownMissingHomeserver,
    WellKnownNotJSON,
    WellKnownNotURL,
    WellKnownUnexpectedStatus,
    WellKnownUnsupportedScheme,
)

__all__ = [
    "IntentError",
    "MatrixConnectionError",
    "MatrixError",
    "MatrixResponseError",
    "CryptoError",
    "DecryptedPayloadError",
    "DecryptionError",
    "DeviceValidationError",
    "DuplicateMessageIndex",
    "EncryptionError",
    "MatchingSessionDecryptionError",
    "MismatchingRoomError",
    "SessionNotFound",
    "SessionShareError",
    "VerificationError",
    "MAlreadyJoined",
    "MatrixBadContent",
    "MatrixBadRequest",
    "MatrixInvalidToken",
    "MatrixRequestError",
    "MatrixStandardRequestError",
    "MatrixUnknownRequestError",
    "MBadJSON",
    "MBadState",
    "MCaptchaInvalid",
    "MCaptchaNeeded",
    "MExclusive",
    "MForbidden",
    "MGuestAccessForbidden",
    "MIncompatibleRoomVersion",
    "MInsufficientPower",
    "MInvalidParam",
    "MInvalidRoomState",
    "MInvalidUsername",
    "MLimitExceeded",
    "MMissingParam",
    "MMissingToken",
    "MNotFound",
    "MNotJoined",
    "MNotJSON",
    "MRoomInUse",
    "MTooLarge",
    "MUnauthorized",
    "MUnknown",
    "MUnknownEndpoint",
    "MUnknownToken",
    "MUnrecognized",
    "MUnsupportedRoomVersion",
    "MUserDeactivated",
    "MUserInUse",
    "make_request_error",
    "standard_error",
    "WellKnownError",
    "WellKnownInvalidVersionsResponse",
    "WellKnownMissingHomeserver",
    "WellKnownNotJSON",
    "WellKnownNotURL",
    "WellKnownUnexpectedStatus",
    "WellKnownUnsupportedScheme",
]



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from typing import Generic, TypeVar

T = TypeVar("T")


class SimpleTemplate(Generic[T]):
    _template: str
    _keyword: str
    _prefix: str
    _suffix: str
    _type: type[T]

    def __init__(
        self, template: str, keyword: str, prefix: str = "", suffix: str = "", type: type[T] = str
    ) -> None:
        self._template = template
        self._keyword = keyword
        index = self._template.find("{%s}" % keyword)
        length = len(keyword) + 2
        self._prefix = prefix + self._template[:index]
        self._suffix = self._template[index + length :] + suffix
        self._type = type

    def format(self, arg: T) -> str:
        return self._template.format(**{self._keyword: arg})

    def format_full(self, arg: T) -> str:
        return f"{self._prefix}{arg}{self._suffix}"

    def parse(self, val: str) -> T | None:
        prefix_ok = val[: len(self._prefix)] == self._prefix
        has_suffix = len(self._suffix) > 0
        suffix_ok = not has_suffix or val[-len(self._suffix) :] == self._suffix
        if prefix_ok and suffix_ok:
            start = len(self._prefix)
            end = -len(self._suffix) if has_suffix else len(val)
            try:
                return self._type(val[start:end])
            except ValueError:
                pass
        return None



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from typing import Any, Awaitable, Callable, NamedTuple, Type
import asyncio
import logging
import time
import traceback

from mautrix.appservice import AppService, IntentAPI
from mautrix.errors import MForbidden
from mautrix.types import EventID, MessageEventContent, RoomID
from mautrix.util import markdown
from mautrix.util.logging import TraceLogger

from ... import bridge as br

command_handlers: dict[str, CommandHandler] = {}
command_aliases: dict[str, CommandHandler] = {}

HelpSection = NamedTuple("HelpSection", name=str, order=int, description=str)
HelpCacheKey = NamedTuple(
    "HelpCacheKey", is_management=bool, is_portal=bool, is_admin=bool, is_logged_in=bool
)

SECTION_GENERAL = HelpSection("General", 0, "")
SECTION_AUTH = HelpSection("Authentication", 10, "")
SECTION_ADMIN = HelpSection("Administration", 50, "")
SECTION_RELAY = HelpSection("Relay mode management", 15, "")


def ensure_trailing_newline(s: str) -> str:
    """Returns the passed string, but with a guaranteed trailing newline."""
    return s + ("" if s[-1] == "\n" else "\n")


class CommandEvent:
    """Holds information about a command issued in a Matrix room.

    When a Matrix command was issued to the bot, CommandEvent will hold
    information regarding the event.

    Attributes:
        room_id: The id of the Matrix room in which the command was issued.
        event_id: The id of the matrix event which contained the command.
        sender: The user who issued the command.
        command: The issued command.
        args: Arguments given with the issued command.
        content: The raw content in the command event.
        portal: The portal the command was sent to.
        is_management: Determines whether the room in which the command was
            issued in is a management room.
        has_bridge_bot: Whether or not the bridge bot is in the room.
    """

    bridge: bridge.Bridge
    az: AppService
    log: TraceLogger
    loop: asyncio.AbstractEventLoop
    config: br.BaseBridgeConfig
    processor: CommandProcessor
    command_prefix: str
    room_id: RoomID
    event_id: EventID
    sender: br.BaseUser
    command: str
    args: list[str]
    content: MessageEventContent
    portal: br.BasePortal | None
    is_management: bool
    has_bridge_bot: bool

    def __init__(
        self,
        processor: CommandProcessor,
        room_id: RoomID,
        event_id: EventID,
        sender: br.BaseUser,
        command: str,
        args: list[str],
        content: MessageEventContent,
        portal: br.BasePortal | None,
        is_management: bool,
        has_bridge_bot: bool,
    ) -> None:
        self.bridge = processor.bridge
        self.az = processor.az
        self.log = processor.log
        self.loop = processor.loop
        self.config = processor.config
        self.processor = processor
        self.command_prefix = processor.command_prefix
        self.room_id = room_id
        self.event_id = event_id
        self.sender = sender
        self.command = command
        self.args = args
        self.content = content
        self.portal = portal
        self.is_management = is_management
        self.has_bridge_bot = has_bridge_bot

    @property
    def is_portal(self) -> bool:
        return self.portal is not None

    async def get_help_key(self) -> HelpCacheKey:
        """
        Get the help cache key for the given CommandEvent.

        Help messages are generated dynamically from the CommandHandlers that have been added so
        that they would only contain relevant commands. The help cache key is tuple-unpacked and
        passed to :meth:`CommandHandler.has_permission` when generating the help page. After the
        first generation, the page is cached using the help cache key.

        If you override this property or :meth:`CommandHandler.has_permission`, make sure to
        override the other too to handle the changes properly.

        When you override this property or otherwise extend CommandEvent, remember to pass the
        extended CommandEvent class when initializing your CommandProcessor.
        """
        return HelpCacheKey(
            is_management=self.is_management,
            is_portal=self.portal is not None,
            is_admin=self.sender.is_admin,
            is_logged_in=await self.sender.is_logged_in(),
        )

    @property
    def print_error_traceback(self) -> bool:
        """
        Whether or not the stack traces of unhandled exceptions during the handling of this command
        should be sent to the user. If false, the error message will simply tell the user to check
        the logs.

        Bridges may want to limit tracebacks to bridge admins.
        """
        return self.sender.is_admin

    @property
    def main_intent(self) -> IntentAPI:
        return self.portal.main_intent if self.portal else self.az.intent

    async def redact(self, reason: str | None = None) -> None:
        """
        Try to redact the command.

        If the redaction fails with M_FORBIDDEN, the error will be logged and ignored.
        """
        try:
            if self.has_bridge_bot:
                await self.az.intent.redact(self.room_id, self.event_id, reason=reason)
            else:
                await self.main_intent.redact(self.room_id, self.event_id, reason=reason)
        except MForbidden as e:
            self.log.warning(f"Failed to redact command {self.command}: {e}")
        except Exception:
            self.log.warning(f"Failed to redact command {self.command}", exc_info=True)

    def reply(
        self, message: str, allow_html: bool = False, render_markdown: bool = True
    ) -> Awaitable[EventID]:
        """Write a reply to the room in which the command was issued.

        Replaces occurences of "$cmdprefix" in the message with the command
        prefix and replaces occurences of "$cmdprefix+sp " with the command
        prefix if the command was not issued in a management room.
        If allow_html and render_markdown are both False, the message will not
        be rendered to html and sending of html is disabled.

        Args:
            message: The message to post in the room.
            allow_html: Escape html in the message or don't render html at all
                if markdown is disabled.
            render_markdown: Use markdown formatting to render the passed
                message to html.

        Returns:
            Handler for the message sending function.
        """
        message = self._replace_command_prefix(message)
        html = self._render_message(
            message, allow_html=allow_html, render_markdown=render_markdown
        )
        if self.has_bridge_bot:
            return self.az.intent.send_notice(self.room_id, message, html=html)
        else:
            return self.main_intent.send_notice(self.room_id, message, html=html)

    async def mark_read(self) -> None:
        """Marks the command as read by the bot."""
        if self.has_bridge_bot:
            await self.az.intent.mark_read(self.room_id, self.event_id)

    def _replace_command_prefix(self, message: str) -> str:
        """Returns the string with the proper command prefix entered."""
        message = message.replace(
            "$cmdprefix+sp ", "" if self.is_management else f"{self.command_prefix} "
        )
        return message.replace("$cmdprefix", self.command_prefix)

    @staticmethod
    def _render_message(message: str, allow_html: bool, render_markdown: bool) -> str | None:
        """Renders the message as HTML.

        Args:
            allow_html: Flag to allow custom HTML in the message.
            render_markdown: If true, markdown styling is applied to the message.

        Returns:
            The message rendered as HTML.
            None is returned if no styled output is required.
        """
        html = ""
        if render_markdown:
            html = markdown.render(message, allow_html=allow_html)
        elif allow_html:
            html = message
        return ensure_trailing_newline(html) if html else None


CommandHandlerFunc = Callable[[CommandEvent], Awaitable[Any]]
IsEnabledForFunc = Callable[[CommandEvent], bool]


class CommandHandler:
    """A command which can be executed from a Matrix room.

    The command manages its permission and help texts.
    When called, it will check the permission of the command event and execute
    the command or, in case of error, report back to the user.

    Attributes:
        management_only: Whether the command can exclusively be issued in a
            management room.
        name: The name of this command.
        help_section: Section of the help in which this command will appear.
    """

    name: str

    management_only: bool
    needs_admin: bool
    needs_auth: bool
    is_enabled_for: IsEnabledForFunc

    _help_text: str
    _help_args: str
    help_section: HelpSection

    def __init__(
        self,
        handler: CommandHandlerFunc,
        management_only: bool,
        name: str,
        help_text: str,
        help_args: str,
        help_section: HelpSection,
        needs_auth: bool,
        needs_admin: bool,
        is_enabled_for: IsEnabledForFunc = lambda _: True,
        **kwargs,
    ) -> None:
        """
        Args:
            handler: The function handling the execution of this command.
            management_only: Whether the command can exclusively be issued
                in a management room.
            needs_auth: Whether the command needs the bridge to be authed already
            needs_admin: Whether the command needs the issuer to be bridge admin
            name: The name of this command.
            help_text: The text displayed in the help for this command.
            help_args: Help text for the arguments of this command.
            help_section: Section of the help in which this command will appear.
        """
        for key, value in kwargs.items():
            setattr(self, key, value)
        self._handler = handler
        self.management_only = management_only
        self.needs_admin = needs_admin
        self.needs_auth = needs_auth
        self.name = name
        self._help_text = help_text
        self._help_args = help_args
        self.help_section = help_section
        self.is_enabled_for = is_enabled_for

    async def get_permission_error(self, evt: CommandEvent) -> str | None:
        """Returns the reason why the command could not be issued.

        Args:
            evt: The event for which to get the error information.

        Returns:
            A string describing the error or None if there was no error.
        """
        if self.management_only and not evt.is_management:
            return (
                f"`{evt.command}` is a restricted command: "
                "you may only run it in management rooms."
            )
        elif self.needs_admin and not evt.sender.is_admin:
            return "That command is limited to bridge administrators."
        elif self.needs_auth and not await evt.sender.is_logged_in():
            return "That command requires you to be logged in."
        return None

    def has_permission(self, key: HelpCacheKey) -> bool:
        """Checks the permission for this command with the given status.

        Args:
            key: The help cache key. See meth:`CommandEvent.get_cache_key`.

        Returns:
            True if a user with the given state is allowed to issue the
            command.
        """
        return (
            (not self.management_only or key.is_management)
            and (not self.needs_admin or key.is_admin)
            and (not self.needs_auth or key.is_logged_in)
        )

    async def __call__(self, evt: CommandEvent) -> Any:
        """Executes the command if evt was issued with proper rights.

        Args:
            evt: The CommandEvent for which to check permissions.

        Returns:
            The result of the command or the error message function.
        """
        error = await self.get_permission_error(evt)
        if error is not None:
            return await evt.reply(error)
        return await self._handler(evt)

    @property
    def has_help(self) -> bool:
        """Returns true if this command has a help text."""
        return bool(self.help_section) and bool(self._help_text)

    @property
    def help(self) -> str:
        """Returns the help text to this command."""
        return f"**{self.name}** {self._help_args} - {self._help_text}"


def command_handler(
    _func: CommandHandlerFunc | None = None,
    *,
    management_only: bool = False,
    name: str | None = None,
    help_text: str = "",
    help_args: str = "",
    help_section: HelpSection = None,
    aliases: list[str] | None = None,
    _handler_class: Type[CommandHandler] = CommandHandler,
    needs_auth: bool = True,
    needs_admin: bool = False,
    is_enabled_for: IsEnabledForFunc = lambda _: True,
    **kwargs,
) -> Callable[[CommandHandlerFunc], CommandHandler]:
    """Decorator to create CommandHandlers"""

    def decorator(func: CommandHandlerFunc) -> CommandHandler:
        actual_name = name or func.__name__.replace("_", "-")
        handler = _handler_class(
            func,
            management_only=management_only,
            name=actual_name,
            help_text=help_text,
            help_args=help_args,
            help_section=help_section,
            needs_auth=needs_auth,
            needs_admin=needs_admin,
            is_enabled_for=is_enabled_for,
            **kwargs,
        )
        command_handlers[handler.name] = handler
        if aliases:
            for alias in aliases:
                command_aliases[alias] = handler
        return handler

    return decorator if _func is None else decorator(_func)


class CommandProcessor:
    """Handles the raw commands issued by a user to the Matrix bot."""

    log: TraceLogger = logging.getLogger("mau.commands")
    az: AppService
    config: br.BaseBridgeConfig
    loop: asyncio.AbstractEventLoop
    event_class: Type[CommandEvent]
    bridge: bridge.Bridge
    _ref_no: int

    def __init__(
        self, bridge: bridge.Bridge, event_class: Type[CommandEvent] = CommandEvent
    ) -> None:
        self.az = bridge.az
        self.config = bridge.config
        self.loop = bridge.loop or asyncio.get_event_loop()
        self.command_prefix = self.config["bridge.command_prefix"]
        self.bridge = bridge
        self.event_class = event_class
        self._ref_no = int(time.time())

    @property
    def ref_no(self) -> int:
        """
        Reference number for a command handling exception to help sysadmins find the error when
        receiving user reports.
        """
        self._ref_no += 1
        return self._ref_no

    @staticmethod
    def _run_handler(
        handler: Callable[[CommandEvent], Awaitable[Any]], evt: CommandEvent
    ) -> Awaitable[Any]:
        return handler(evt)

    async def handle(
        self,
        room_id: RoomID,
        event_id: EventID,
        sender: br.BaseUser,
        command: str,
        args: list[str],
        content: MessageEventContent,
        portal: br.BasePortal | None,
        is_management: bool,
        has_bridge_bot: bool,
    ) -> None:
        """Handles the raw commands issued by a user to the Matrix bot.

        If the command is not known, it might be a followup command and is
        delegated to a command handler registered for that purpose in the
        senders command_status as "next".

        Args:
            room_id: ID of the Matrix room in which the command was issued.
            event_id: ID of the event by which the command was issued.
            sender: The sender who issued the command.
            command: The issued command, case insensitive.
            args: Arguments given with the command.
            content: The raw content in the command event.
            portal: The portal the command was sent to.
            is_management: Whether the room is a management room.
            has_bridge_bot: Whether or not the bridge bot is in the room.

        Returns:
            The result of the error message function or None if no error
            occured. Unknown and delegated commands do not count as errors.
        """
        if not command_handlers or "unknown-command" not in command_handlers:
            raise ValueError("command_handlers are not properly initialized.")

        evt = self.event_class(
            processor=self,
            room_id=room_id,
            event_id=event_id,
            sender=sender,
            command=command,
            args=args,
            content=content,
            portal=portal,
            is_management=is_management,
            has_bridge_bot=has_bridge_bot,
        )
        orig_command = command
        command = command.lower()

        handler = command_handlers.get(command, command_aliases.get(command))
        if handler is None or not handler.is_enabled_for(evt):
            if sender.command_status and "next" in sender.command_status:
                args.insert(0, orig_command)
                evt.command = ""
                handler = sender.command_status["next"]
            else:
                handler = command_handlers["unknown-command"]

        try:
            await self._run_handler(handler, evt)
        except Exception:
            ref_no = self.ref_no
            self.log.exception(
                "Unhandled error while handling command "
                f"{evt.command} {' '.join(args)} from {sender.mxid} (ref: {ref_no})"
            )
            if evt.print_error_traceback:
                await evt.reply(
                    "Unhandled error while handling command:\n\n"
                    "```traceback\n"
                    f"{traceback.format_exc()}"
                    "```"
                )
            else:
                await evt.reply(
                    "Unhandled error while handling command. "
                    f"Check logs for more details (ref: {ref_no})."
                )
            raise
        return None



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Dict, NewType, Optional, Union
from enum import IntEnum
import warnings

from attr import dataclass

from ..primitive import JSON, DeviceID, IdentityKey, SessionID
from ..util import ExtensibleEnum, Obj, Serializable, SerializableAttrs, deserializer, field
from .base import BaseRoomEvent, BaseUnsigned
from .message import RelatesTo


class EncryptionAlgorithm(ExtensibleEnum):
    OLM_V1: "EncryptionAlgorithm" = "m.olm.v1.curve25519-aes-sha2"
    MEGOLM_V1: "EncryptionAlgorithm" = "m.megolm.v1.aes-sha2"


class EncryptionKeyAlgorithm(ExtensibleEnum):
    CURVE25519: "EncryptionKeyAlgorithm" = "curve25519"
    ED25519: "EncryptionKeyAlgorithm" = "ed25519"
    SIGNED_CURVE25519: "EncryptionKeyAlgorithm" = "signed_curve25519"


@dataclass(frozen=True)
class KeyID(Serializable):
    algorithm: EncryptionKeyAlgorithm
    key_id: str

    def serialize(self) -> JSON:
        return str(self)

    @classmethod
    def deserialize(cls, raw: JSON) -> "KeyID":
        assert isinstance(raw, str), "key IDs must be strings"
        alg, key_id = raw.split(":", 1)
        return cls(EncryptionKeyAlgorithm(alg), key_id)

    def __str__(self) -> str:
        return f"{self.algorithm.value}:{self.key_id}"


class OlmMsgType(Serializable, IntEnum):
    PREKEY = 0
    MESSAGE = 1

    def serialize(self) -> JSON:
        return self.value

    @classmethod
    def deserialize(cls, raw: JSON) -> "OlmMsgType":
        return cls(raw)


@dataclass
class OlmCiphertext(SerializableAttrs):
    body: str
    type: OlmMsgType


@dataclass
class EncryptedOlmEventContent(SerializableAttrs):
    ciphertext: Dict[str, OlmCiphertext]
    sender_key: IdentityKey
    algorithm: EncryptionAlgorithm = EncryptionAlgorithm.OLM_V1


@dataclass
class EncryptedMegolmEventContent(SerializableAttrs):
    """The content of an m.room.encrypted event"""

    ciphertext: str
    session_id: SessionID
    algorithm: EncryptionAlgorithm = EncryptionAlgorithm.MEGOLM_V1

    _sender_key: Optional[IdentityKey] = field(default=None, json="sender_key")
    _device_id: Optional[DeviceID] = field(default=None, json="device_id")
    _relates_to: Optional[RelatesTo] = field(default=None, json="m.relates_to")

    @property
    def sender_key(self) -> Optional[IdentityKey]:
        """
        .. deprecated:: 0.17.0
            Matrix v1.3 deprecated the device_id and sender_key fields in megolm events.
        """
        warnings.warn(
            "The sender_key field in Megolm events was deprecated in Matrix 1.3",
            DeprecationWarning,
        )
        return self._sender_key

    @property
    def device_id(self) -> Optional[DeviceID]:
        """
        .. deprecated:: 0.17.0
            Matrix v1.3 deprecated the device_id and sender_key fields in megolm events.
        """
        warnings.warn(
            "The sender_key field in Megolm events was deprecated in Matrix 1.3",
            DeprecationWarning,
        )
        return self._device_id

    @property
    def relates_to(self) -> RelatesTo:
        if self._relates_to is None:
            self._relates_to = RelatesTo()
        return self._relates_to

    @relates_to.setter
    def relates_to(self, relates_to: RelatesTo) -> None:
        self._relates_to = relates_to


EncryptedEventContent = NewType(
    "EncryptedEventContent", Union[EncryptedOlmEventContent, EncryptedMegolmEventContent]
)


@deserializer(EncryptedEventContent)
def deserialize_encrypted(data: JSON) -> Union[EncryptedEventContent, Obj]:
    alg = data.get("algorithm", None)
    if alg == EncryptionAlgorithm.MEGOLM_V1.value:
        return EncryptedMegolmEventContent.deserialize(data)
    elif alg == EncryptionAlgorithm.OLM_V1.value:
        return EncryptedOlmEventContent.deserialize(data)
    return Obj(**data)


setattr(EncryptedEventContent, "deserialize", deserialize_encrypted)


@dataclass
class EncryptedEvent(BaseRoomEvent, SerializableAttrs):
    """A m.room.encrypted event"""

    content: EncryptedEventContent
    _unsigned: Optional[BaseUnsigned] = field(default=None, json="unsigned")

    @property
    def unsigned(self) -> BaseUnsigned:
        if not self._unsigned:
            self._unsigned = BaseUnsigned()
        return self._unsigned

    @unsigned.setter
    def unsigned(self, value: BaseUnsigned) -> None:
        self._unsigned = value



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from typing import Type
import json

from sqlalchemy import Boolean, Column, Text, types

from mautrix.types import (
    PowerLevelStateEventContent as PowerLevels,
    RoomEncryptionStateEventContent as EncryptionInfo,
    RoomID,
    Serializable,
)
from mautrix.util.db import Base


class SerializableType(types.TypeDecorator):
    impl = types.Text

    def __init__(self, python_type: Type[Serializable], *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._python_type = python_type

    @property
    def python_type(self) -> Type[Serializable]:
        return self._python_type

    def process_bind_param(self, value: Serializable, dialect) -> str | None:
        if value is not None:
            return json.dumps(value.serialize() if isinstance(value, Serializable) else value)
        return None

    def process_result_value(self, value: str, dialect) -> Serializable | None:
        if value is not None:
            return self.python_type.deserialize(json.loads(value))
        return None

    def process_literal_param(self, value, dialect):
        return value


class RoomState(Base):
    __tablename__ = "mx_room_state"

    room_id: RoomID = Column(Text, primary_key=True)
    is_encrypted: bool = Column(Boolean, nullable=True)
    has_full_member_list: bool = Column(Boolean, nullable=True)
    encryption: EncryptionInfo = Column(SerializableType(EncryptionInfo), nullable=True)
    power_levels: PowerLevels = Column(SerializableType(PowerLevels), nullable=True)

    @property
    def has_power_levels(self) -> bool:
        return bool(self.power_levels)

    @property
    def has_encryption_info(self) -> bool:
        return self.is_encrypted is not None

    @classmethod
    def get(cls, room_id: RoomID) -> RoomState | None:
        return cls._select_one_or_none(cls.c.room_id == room_id)



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import List, Optional, Union

from attr import dataclass
import attr

from .primitive import JSON, RoomID, UserID
from .util import ExtensibleEnum, SerializableAttrs, deserializer

PushRuleID = Union[RoomID, UserID, str]


class PushActionType(ExtensibleEnum):
    NOTIFY = "notify"
    DONT_NOTIFY = "dont_notify"
    COALESCE = "coalesce"


@dataclass
class PushActionDict(SerializableAttrs):
    set_tweak: Optional[str] = None
    value: Optional[str] = None


PushAction = Union[PushActionDict, PushActionType]


@deserializer(PushAction)
def deserialize_push_action(data: JSON) -> PushAction:
    if isinstance(data, str):
        return PushActionType(data)
    else:
        return PushActionDict.deserialize(data)


class PushOperator(ExtensibleEnum):
    EQUALS = "=="
    LESS_THAN = "<"
    GREATER_THAN = ">"
    LESS_THAN_OR_EQUAL = "<="
    GREATER_THAN_OR_EQUAL = ">="


class PushRuleScope(ExtensibleEnum):
    GLOBAL = "global"


class PushConditionKind(ExtensibleEnum):
    EVENT_MATCH = "event_match"
    CONTAINS_DISPLAY_NAME = "contains_display_name"
    ROOM_MEMBER_COUNT = "room_member_count"
    SENDER_NOTIFICATION_PERMISSION = "sender_notification_permission"


class PushRuleKind(ExtensibleEnum):
    OVERRIDE = "override"
    SENDER = "sender"
    ROOM = "room"
    CONTENT = "content"
    UNDERRIDE = "underride"


@dataclass
class PushCondition(SerializableAttrs):
    kind: PushConditionKind
    key: Optional[str] = None
    pattern: Optional[str] = None
    operator: PushOperator = attr.ib(
        default=PushOperator.EQUALS, metadata={"json": "is", "omitdefault": True}
    )


@dataclass
class PushRule(SerializableAttrs):
    rule_id: PushRuleID
    default: bool
    enabled: bool
    actions: List[PushAction]
    conditions: List[PushCondition] = attr.ib(factory=lambda: [])
    pattern: Optional[str] = None



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

import json
import pkgutil

import aiohttp

EMOJI_VAR_URL = "https://www.unicode.org/Public/14.0.0/ucd/emoji/emoji-variation-sequences.txt"


def read_data() -> dict[str, str]:
    """
    Get the list of emoji that need a variation selector. This loads the local data file that was
    previously generated from the Unicode spec data files.

    Returns:
        A dict from hex to the emoji string (you have to bring the variation selectors yourself).
    """
    return json.loads(pkgutil.get_data("mautrix.util", "variation_selector.json"))


async def fetch_data() -> dict[str, str]:
    """
    Generate the list of emoji that need a variation selector from the Unicode spec data files.

    Returns:
        A dict from hex to the emoji string (you have to bring the variation selectors yourself).
    """
    async with aiohttp.ClientSession() as sess, sess.get(EMOJI_VAR_URL) as resp:
        data = await resp.text()
    emojis = {}
    for line in data.split("\n"):
        if "emoji style" in line:
            emoji_hex = line.split(" ", 1)[0]
            emojis[emoji_hex] = rf"\U{emoji_hex:>08}".encode("ascii").decode("unicode-escape")
    return emojis


if __name__ == "__main__":
    import asyncio
    import sys

    import pkg_resources

    path = pkg_resources.resource_filename("mautrix.util", "variation_selector.json")
    emojis = asyncio.run(fetch_data())
    with open(path, "w") as file:
        json.dump(emojis, file, indent="    ", ensure_ascii=False)
        file.write("\n")
    print(f"Wrote {len(emojis)} emojis to {path}")
    sys.exit(0)

VARIATION_SELECTOR_16 = "\ufe0f"
ADD_VARIATION_TRANSLATION = str.maketrans(
    {ord(emoji): f"{emoji}{VARIATION_SELECTOR_16}" for emoji in read_data().values()}
)
SKIN_TONE_MODIFIERS = ("\U0001F3FB", "\U0001F3FC", "\U0001F3FD", "\U0001F3FE", "\U0001F3FF")
SKIN_TONE_REPLACEMENTS = {f"{VARIATION_SELECTOR_16}{mod}": mod for mod in SKIN_TONE_MODIFIERS}
VARIATION_SELECTOR_REPLACEMENTS = {
    **SKIN_TONE_REPLACEMENTS,
    "\U0001F408\ufe0f\u200d\u2b1b\ufe0f": "\U0001F408\u200d\u2b1b",
}


def add(val: str) -> str:
    r"""
    Add emoji variation selectors (16) to all emojis that have multiple forms in the given string.
    This will remove all variation selectors first to make sure it doesn't add duplicates.

    .. versionadded:: 0.12.5

    Examples:
        >>> from mautrix.util import variation_selector
        >>> variation_selector.add("\U0001f44d")
        "\U0001f44d\ufe0f"
        >>> variation_selector.add("\U0001f44d\ufe0f")
        "\U0001f44d\ufe0f"
        >>> variation_selector.add("4\u20e3")
        "4\ufe0f\u20e3"
        >>> variation_selector.add("\U0001f9d0")
        "\U0001f9d0"

    Args:
        val: The string to add variation selectors to.

    Returns:
        The string with variation selectors added.
    """
    added = remove(val).translate(ADD_VARIATION_TRANSLATION)
    for invalid_selector, replacement in VARIATION_SELECTOR_REPLACEMENTS.items():
        added = added.replace(invalid_selector, replacement)
    return added


def remove(val: str) -> str:
    """
    Remove all emoji variation selectors in the given string.

    .. versionadded:: 0.12.5

    Args:
        val: The string to remove variation selectors from.

    Returns:
        The string with variation selectors removed.
    """
    return val.replace(VARIATION_SELECTOR_16, "")


__all__ = ["add", "remove", "read_data", "fetch_data"]



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Optional

from attr import dataclass

from ..primitive import EventID
from ..util import SerializableAttrs, SerializableEnum, field
from .base import BaseRoomEvent
from .message import RelatesTo


class MessageStatusReason(SerializableEnum):
    GENERIC_ERROR = "m.event_not_handled"
    UNSUPPORTED = "com.beeper.unsupported_event"
    UNDECRYPTABLE = "com.beeper.undecryptable_event"
    TOO_OLD = "m.event_too_old"
    NETWORK_ERROR = "m.foreign_network_error"
    NO_PERMISSION = "m.no_permission"

    @property
    def checkpoint_status(self):
        from mautrix.util.message_send_checkpoint import MessageSendCheckpointStatus

        if self == MessageStatusReason.UNSUPPORTED:
            return MessageSendCheckpointStatus.UNSUPPORTED
        elif self == MessageStatusReason.TOO_OLD:
            return MessageSendCheckpointStatus.TIMEOUT
        return MessageSendCheckpointStatus.PERM_FAILURE


class MessageStatus(SerializableEnum):
    SUCCESS = "SUCCESS"
    PENDING = "PENDING"
    RETRIABLE = "FAIL_RETRIABLE"
    FAIL = "FAIL_PERMANENT"


@dataclass(kw_only=True)
class BeeperMessageStatusEventContent(SerializableAttrs):
    relates_to: RelatesTo = field(json="m.relates_to")
    network: str = ""
    status: Optional[MessageStatus] = None

    reason: Optional[MessageStatusReason] = None
    error: Optional[str] = None
    message: Optional[str] = None

    success: Optional[bool] = None
    still_working: Optional[bool] = None
    can_retry: Optional[bool] = None
    is_certain: Optional[bool] = None

    last_retry: Optional[EventID] = None

    def fill_legacy_booleans(self) -> None:
        self.success = self.status == MessageStatus.SUCCESS
        if not self.success:
            self.still_working = self.status == MessageStatus.PENDING
            self.can_retry = self.status in (MessageStatus.PENDING, MessageStatus.RETRIABLE)


@dataclass
class BeeperMessageStatusEvent(BaseRoomEvent, SerializableAttrs):
    content: BeeperMessageStatusEventContent



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from typing import Any, NamedTuple
import json

from mautrix.types import (
    Member,
    Membership,
    MemberStateEventContent,
    PowerLevelStateEventContent,
    RoomEncryptionStateEventContent,
    RoomID,
    Serializable,
    UserID,
)
from mautrix.util.async_db import Database, Scheme

from ..abstract import StateStore
from .upgrade import upgrade_table


class RoomState(NamedTuple):
    is_encrypted: bool
    has_full_member_list: bool
    encryption: RoomEncryptionStateEventContent
    power_levels: PowerLevelStateEventContent


class PgStateStore(StateStore):
    upgrade_table = upgrade_table

    db: Database

    def __init__(self, db: Database) -> None:
        self.db = db

    async def get_member(self, room_id: RoomID, user_id: UserID) -> Member | None:
        res = await self.db.fetchrow(
            "SELECT membership, displayname, avatar_url "
            "FROM mx_user_profile WHERE room_id=$1 AND user_id=$2",
            room_id,
            user_id,
        )
        if res is None:
            return None
        return Member(
            membership=Membership.deserialize(res["membership"]),
            displayname=res["displayname"],
            avatar_url=res["avatar_url"],
        )

    async def set_member(
        self, room_id: RoomID, user_id: UserID, member: Member | MemberStateEventContent
    ) -> None:
        q = (
            "INSERT INTO mx_user_profile (room_id, user_id, membership, displayname, avatar_url) "
            "VALUES ($1, $2, $3, $4, $5)"
            "ON CONFLICT (room_id, user_id) DO UPDATE SET membership=$3, displayname=$4,"
            "                                             avatar_url=$5"
        )
        await self.db.execute(
            q, room_id, user_id, member.membership.value, member.displayname, member.avatar_url
        )

    async def set_membership(
        self, room_id: RoomID, user_id: UserID, membership: Membership
    ) -> None:
        q = (
            "INSERT INTO mx_user_profile (room_id, user_id, membership) VALUES ($1, $2, $3) "
            "ON CONFLICT (room_id, user_id) DO UPDATE SET membership=$3"
        )
        await self.db.execute(q, room_id, user_id, membership.value)

    async def get_members(
        self,
        room_id: RoomID,
        memberships: tuple[Membership, ...] = (Membership.JOIN, Membership.INVITE),
    ) -> list[UserID]:
        membership_values = [membership.value for membership in memberships]
        if self.db.scheme in (Scheme.POSTGRES, Scheme.COCKROACH):
            q = "SELECT user_id FROM mx_user_profile WHERE room_id=$1 AND membership=ANY($2)"
            res = await self.db.fetch(q, room_id, membership_values)
        else:
            membership_placeholders = ("?," * len(memberships)).rstrip(",")
            q = (
                "SELECT user_id FROM mx_user_profile "
                f"WHERE room_id=? AND membership IN ({membership_placeholders})"
            )
            res = await self.db.fetch(q, room_id, *membership_values)
        return [profile["user_id"] for profile in res]

    async def get_member_profiles(
        self,
        room_id: RoomID,
        memberships: tuple[Membership, ...] = (Membership.JOIN, Membership.INVITE),
    ) -> dict[UserID, Member]:
        membership_values = [membership.value for membership in memberships]
        if self.db.scheme in (Scheme.POSTGRES, Scheme.COCKROACH):
            q = (
                "SELECT user_id, membership, displayname, avatar_url FROM mx_user_profile "
                "WHERE room_id=$1 AND membership=ANY($2)"
            )
            res = await self.db.fetch(q, room_id, membership_values)
        else:
            membership_placeholders = ("?," * len(memberships)).rstrip(",")
            q = (
                "SELECT user_id, membership, displayname, avatar_url FROM mx_user_profile "
                f"WHERE room_id=? AND membership IN ({membership_placeholders})"
            )
            res = await self.db.fetch(q, room_id, *membership_values)
        return {profile["user_id"]: Member.deserialize(profile) for profile in res}

    async def get_members_filtered(
        self,
        room_id: RoomID,
        not_prefix: str,
        not_suffix: str,
        not_id: str,
        memberships: tuple[Membership, ...] = (Membership.JOIN, Membership.INVITE),
    ) -> list[UserID]:
        not_like = f"{not_prefix}%{not_suffix}"
        membership_values = [membership.value for membership in memberships]
        if self.db.scheme in (Scheme.POSTGRES, Scheme.COCKROACH):
            q = (
                "SELECT user_id FROM mx_user_profile "
                "WHERE room_id=$1 AND membership=ANY($2)"
                "AND user_id != $3 AND user_id NOT LIKE $4"
            )
            res = await self.db.fetch(q, room_id, membership_values, not_id, not_like)
        else:
            membership_placeholders = ("?," * len(memberships)).rstrip(",")
            q = (
                "SELECT user_id FROM mx_user_profile "
                f"WHERE room_id=? AND membership IN ({membership_placeholders})"
                "AND user_id != ? AND user_id NOT LIKE ?"
            )
            res = await self.db.fetch(q, room_id, *membership_values, not_id, not_like)
        return [profile["user_id"] for profile in res]

    async def set_members(
        self,
        room_id: RoomID,
        members: dict[UserID, Member | MemberStateEventContent],
        only_membership: Membership | None = None,
    ) -> None:
        columns = ["room_id", "user_id", "membership", "displayname", "avatar_url"]
        records = [
            (room_id, user_id, str(member.membership), member.displayname, member.avatar_url)
            for user_id, member in members.items()
        ]
        async with self.db.acquire() as conn, conn.transaction():
            del_q = "DELETE FROM mx_user_profile WHERE room_id=$1"
            if only_membership is None:
                await conn.execute(del_q, room_id)
            elif self.db.scheme in (Scheme.POSTGRES, Scheme.COCKROACH):
                del_q = f"{del_q} AND (membership=$2 OR user_id = ANY($3))"
                await conn.execute(del_q, room_id, only_membership.value, list(members.keys()))
            else:
                member_placeholders = ("?," * len(members)).rstrip(",")
                del_q = f"{del_q} AND (membership=? OR user_id IN ({member_placeholders}))"
                await conn.execute(del_q, room_id, only_membership.value, *members.keys())

            if self.db.scheme == Scheme.POSTGRES:
                await conn.copy_records_to_table(
                    "mx_user_profile", records=records, columns=columns
                )
            else:
                q = (
                    "INSERT INTO mx_user_profile (room_id, user_id, membership, "
                    "displayname, avatar_url) VALUES ($1, $2, $3, $4, $5)"
                )
                await conn.executemany(q, records)

            if not only_membership or only_membership == Membership.JOIN:
                await conn.execute(
                    "UPDATE mx_room_state SET has_full_member_list=true WHERE room_id=$1",
                    room_id,
                )

    async def find_shared_rooms(self, user_id: UserID) -> list[RoomID]:
        q = (
            "SELECT mx_user_profile.room_id FROM mx_user_profile "
            "LEFT JOIN mx_room_state ON mx_room_state.room_id=mx_user_profile.room_id "
            "WHERE user_id=$1 AND mx_room_state.is_encrypted=true"
        )
        rows = await self.db.fetch(q, user_id)
        return [row["room_id"] for row in rows]

    async def has_full_member_list(self, room_id: RoomID) -> bool:
        return bool(
            await self.db.fetchval(
                "SELECT has_full_member_list FROM mx_room_state WHERE room_id=$1", room_id
            )
        )

    async def has_power_levels_cached(self, room_id: RoomID) -> bool:
        return bool(
            await self.db.fetchval(
                "SELECT power_levels IS NOT NULL FROM mx_room_state WHERE room_id=$1", room_id
            )
        )

    async def get_power_levels(self, room_id: RoomID) -> PowerLevelStateEventContent | None:
        power_levels_json = await self.db.fetchval(
            "SELECT power_levels FROM mx_room_state WHERE room_id=$1", room_id
        )
        if power_levels_json is None:
            return None
        return PowerLevelStateEventContent.parse_json(power_levels_json)

    async def set_power_levels(
        self, room_id: RoomID, content: PowerLevelStateEventContent | dict[str, Any]
    ) -> None:
        await self.db.execute(
            "INSERT INTO mx_room_state (room_id, power_levels) VALUES ($1, $2) "
            "ON CONFLICT (room_id) DO UPDATE SET power_levels=$2",
            room_id,
            json.dumps(content.serialize() if isinstance(content, Serializable) else content),
        )

    async def has_encryption_info_cached(self, room_id: RoomID) -> bool:
        return bool(
            await self.db.fetchval(
                "SELECT encryption IS NULL FROM mx_room_state WHERE room_id=$1", room_id
            )
        )

    async def is_encrypted(self, room_id: RoomID) -> bool | None:
        return await self.db.fetchval(
            "SELECT is_encrypted FROM mx_room_state WHERE room_id=$1", room_id
        )

    async def get_encryption_info(self, room_id: RoomID) -> RoomEncryptionStateEventContent | None:
        row = await self.db.fetchrow(
            "SELECT is_encrypted, encryption FROM mx_room_state WHERE room_id=$1", room_id
        )
        if row is None or not row["is_encrypted"]:
            return None
        return RoomEncryptionStateEventContent.parse_json(row["encryption"])

    async def set_encryption_info(
        self, room_id: RoomID, content: RoomEncryptionStateEventContent | dict[str, Any]
    ) -> None:
        q = (
            "INSERT INTO mx_room_state (room_id, is_encrypted, encryption) VALUES ($1, true, $2) "
            "ON CONFLICT (room_id) DO UPDATE SET is_encrypted=true, encryption=$2"
        )
        await self.db.execute(
            q,
            room_id,
            json.dumps(content.serialize() if isinstance(content, Serializable) else content),
        )



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from typing import IO, Any, Protocol
from abc import ABC, abstractmethod
from pathlib import Path
import json
import pickle
import time


class Filer(Protocol):
    def dump(self, obj: Any, file: IO) -> None:
        pass

    def load(self, file: IO) -> Any:
        pass


class FileStore(ABC):
    path: str | Path | IO
    filer: Filer
    binary: bool
    save_interval: float
    _last_save: float

    def __init__(
        self,
        path: str | Path | IO,
        filer: Filer | None = None,
        binary: bool = True,
        save_interval: float = 60.0,
    ) -> None:
        self.path = path
        self.filer = filer or (pickle if binary else json)
        self.binary = binary
        self.save_interval = save_interval
        self._last_save = time.monotonic()

    @abstractmethod
    def serialize(self) -> Any:
        pass

    @abstractmethod
    def deserialize(self, data: Any) -> None:
        pass

    def _save(self) -> None:
        if isinstance(self.path, IO):
            file = self.path
            close = False
        else:
            file = open(self.path, "wb" if self.binary else "w")
            close = True
        try:
            self.filer.dump(self.serialize(), file)
        finally:
            if close:
                file.close()

    def _load(self) -> None:
        if isinstance(self.path, IO):
            file = self.path
            close = False
        else:
            try:
                file = open(self.path, "rb" if self.binary else "r")
            except FileNotFoundError:
                return
            close = True
        try:
            self.deserialize(self.filer.load(file))
        finally:
            if close:
                file.close()

    async def flush(self) -> None:
        self._save()

    async def open(self) -> None:
        self._load()

    def _time_limited_flush(self) -> None:
        if self._last_save + self.save_interval < time.monotonic():
            self._save()
            self._last_save = time.monotonic()



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import List, NamedTuple, NewType, Optional
from enum import Enum

from attr import dataclass
import attr

from .event import Event, StateEvent
from .primitive import BatchID, ContentURI, EventID, RoomAlias, RoomID, SyncToken, UserID
from .util import SerializableAttrs


@dataclass
class DeviceLists(SerializableAttrs):
    changed: List[UserID] = attr.ib(factory=lambda: [])
    left: List[UserID] = attr.ib(factory=lambda: [])

    def __bool__(self) -> bool:
        return bool(self.changed or self.left)


@dataclass
class DeviceOTKCount(SerializableAttrs):
    signed_curve25519: int = 0
    curve25519: int = 0


class RoomCreatePreset(Enum):
    """
    Room creation preset, as specified in the `createRoom endpoint`_

    .. _createRoom endpoint:
        https://spec.matrix.org/v1.2/client-server-api/#post_matrixclientv3createroom
    """

    PRIVATE = "private_chat"
    TRUSTED_PRIVATE = "trusted_private_chat"
    PUBLIC = "public_chat"


class RoomDirectoryVisibility(Enum):
    """
    Room directory visibility, as specified in the `createRoom endpoint`_

    .. _createRoom endpoint:
        https://spec.matrix.org/v1.2/client-server-api/#post_matrixclientv3createroom
    """

    PRIVATE = "private"
    PUBLIC = "public"


class PaginationDirection(Enum):
    """Pagination direction used in various endpoints that support pagination."""

    FORWARD = "f"
    BACKWARD = "b"


@dataclass
class RoomAliasInfo(SerializableAttrs):
    """
    Room alias query result, as specified in the `alias resolve endpoint`_

    .. _alias resolve endpoint:
        https://spec.matrix.org/v1.2/client-server-api/#get_matrixclientv3directoryroomroomalias
    """

    room_id: RoomID = None
    """The room ID for this room alias."""

    servers: List[str] = None
    """A list of servers that are aware of this room alias."""


DirectoryPaginationToken = NewType("DirectoryPaginationToken", str)


@dataclass
class PublicRoomInfo(SerializableAttrs):
    room_id: RoomID

    num_joined_members: int

    world_readable: bool
    guests_can_join: bool

    name: str = None
    topic: str = None
    avatar_url: ContentURI = None

    aliases: List[RoomAlias] = None
    canonical_alias: RoomAlias = None


@dataclass
class RoomDirectoryResponse(SerializableAttrs):
    chunk: List[PublicRoomInfo]
    next_batch: DirectoryPaginationToken = None
    prev_batch: DirectoryPaginationToken = None
    total_room_count_estimate: int = None


PaginatedMessages = NamedTuple(
    "PaginatedMessages", start=SyncToken, end=SyncToken, events=List[Event]
)


@dataclass
class EventContext(SerializableAttrs):
    end: SyncToken
    start: SyncToken
    event: Event
    events_after: List[Event]
    events_before: List[Event]
    state: List[StateEvent]


@dataclass
class BatchSendResponse(SerializableAttrs):
    state_event_ids: List[EventID]
    event_ids: List[EventID]

    insertion_event_id: EventID
    batch_event_id: EventID
    next_batch_id: BatchID
    base_insertion_event_id: Optional[EventID] = None



import setuptools

from mautrix import __version__

encryption_dependencies = ["python-olm", "unpaddedbase64", "pycryptodome"]
test_dependencies = ["aiosqlite", "sqlalchemy", "asyncpg", *encryption_dependencies]

setuptools.setup(
    name="mautrix",
    version=__version__,
    url="https://github.com/mautrix/python",
    project_urls={
        "Changelog": "https://github.com/mautrix/python/blob/master/CHANGELOG.md",
    },

    author="Tulir Asokan",
    author_email="tulir@maunium.net",

    description="A Python 3 asyncio Matrix framework.",
    long_description=open("README.rst").read(),

    packages=setuptools.find_packages(),

    install_requires=[
        "aiohttp>=3,<4",
        "attrs>=18.1.0",
        "yarl>=1.5,<2",
    ],
    extras_require={
        "detect_mimetype": ["python-magic>=0.4.15,<0.5"],
        "lint": ["black==22.1.0", "isort"],
        "test": ["pytest", "pytest-asyncio", *test_dependencies],
        "encryption": encryption_dependencies,
    },
    tests_require=test_dependencies,
    python_requires="~=3.8",

    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)",
        "Topic :: Communications :: Chat",
        "Framework :: AsyncIO",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],

    package_data={
        "mautrix": ["py.typed"],
        "mautrix.types.event": ["type.pyi"],
        "mautrix.util": ["opt_prometheus.pyi", "variation_selector.json"],
        "mautrix.util.formatter": ["html_reader.pyi"],
    },
)



# From https://github.com/LonamiWebs/Telethon/blob/v1.24.0/telethon/helpers.py#L38-L62
# Copyright (c) LonamiWebs, MIT license

import struct


def add(text: str) -> str:
    """
    Add surrogate pairs to characters in the text. This makes the indices match how most platforms
    calculate string length when formatting texts using offset-based entities.

    Args:
        text: The text to add surrogate pairs to.

    Returns:
        The text with surrogate pairs.
    """
    return "".join(
        "".join(chr(y) for y in struct.unpack("<HH", x.encode("utf-16le")))
        if (0x10000 <= ord(x) <= 0x10FFFF)
        else x
        for x in text
    )


def remove(text: str) -> str:
    """
    Remove surrogate pairs from text. This does the opposite of :func:`add`.

    Args:
        text: The text with surrogate pairs.

    Returns:
        The text without surrogate pairs.
    """
    return text.encode("utf-16", "surrogatepass").decode("utf-16")


def is_within(text: str, index: int, *, length: int = None) -> bool:
    """
    Returns:
        `True` if ``index`` is within a surrogate (before and after it, not at!).
    """
    if length is None:
        length = len(text)

    return (
        1 < index < len(text)
        and "\ud800" <= text[index - 1] <= "\udfff"  # in bounds
        and "\ud800" <= text[index] <= "\udfff"  # previous is  # current is
    )


__all__ = ["add", "remove"]



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from mautrix.client.state_store.sqlalchemy import SQLStateStore as SQLClientStateStore

from .memory import ASStateStore


class SQLASStateStore(SQLClientStateStore, ASStateStore):
    def __init__(self) -> None:
        SQLClientStateStore.__init__(self)
        ASStateStore.__init__(self)



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.


def _pluralize(count: int, singular: str) -> str:
    return singular if count == 1 else f"{singular}s"


def _include_if_positive(count: int, word: str) -> str:
    return f"{count} {_pluralize(count, word)}" if count > 0 else ""


def format_duration(seconds: int) -> str:
    """
    Format seconds as a simple duration in weeks/days/hours/minutes/seconds.

    Args:
        seconds: The number of seconds as an integer. Must be positive.

    Returns:
        The formatted duration.

    Examples:
        >>> from mautrix.util.format_duration import format_duration
        >>> format_duration(1234)
        '20 minutes and 34 seconds'
        >>> format_duration(987654)
        '1 week, 4 days, 10 hours, 20 minutes and 54 seconds'
        >>> format_duration(60)
        '1 minute'

    Raises:
        ValueError: if the duration is not positive.
    """
    if seconds <= 0:
        raise ValueError("format_duration only accepts positive values")
    minutes, seconds = divmod(seconds, 60)
    hours, minutes = divmod(minutes, 60)
    days, hours = divmod(hours, 24)
    weeks, days = divmod(days, 7)

    parts = [
        _include_if_positive(weeks, "week"),
        _include_if_positive(days, "day"),
        _include_if_positive(hours, "hour"),
        _include_if_positive(minutes, "minute"),
        _include_if_positive(seconds, "second"),
    ]
    parts = [part for part in parts if part]
    if len(parts) > 2:
        parts = [", ".join(parts[:-1]), parts[-1]]
    return " and ".join(parts)



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

import logging

from mautrix.util.async_db import Connection, Scheme, UpgradeTable

upgrade_table = UpgradeTable(
    version_table_name="crypto_version",
    database_name="crypto store",
    log=logging.getLogger("mau.crypto.db.upgrade"),
)


@upgrade_table.register(description="Latest revision", upgrades_to=5)
async def upgrade_blank_to_v4(conn: Connection) -> None:
    await conn.execute(
        """CREATE TABLE IF NOT EXISTS crypto_account (
            account_id TEXT PRIMARY KEY,
            device_id  TEXT,
            shared     BOOLEAN      NOT NULL,
            sync_token TEXT         NOT NULL,
            account    bytea        NOT NULL
        )"""
    )
    await conn.execute(
        """CREATE TABLE IF NOT EXISTS crypto_message_index (
            sender_key CHAR(43),
            session_id CHAR(43),
            "index"    INTEGER,
            event_id   TEXT   NOT NULL,
            timestamp  BIGINT NOT NULL,
            PRIMARY KEY (sender_key, session_id, "index")
        )"""
    )
    await conn.execute(
        """CREATE TABLE IF NOT EXISTS crypto_tracked_user (
            user_id TEXT PRIMARY KEY
        )"""
    )
    await conn.execute(
        """CREATE TABLE IF NOT EXISTS crypto_device (
            user_id      TEXT,
            device_id    TEXT,
            identity_key CHAR(43) NOT NULL,
            signing_key  CHAR(43) NOT NULL,
            trust        SMALLINT NOT NULL,
            deleted      BOOLEAN  NOT NULL,
            name         TEXT     NOT NULL,
            PRIMARY KEY (user_id, device_id)
        )"""
    )
    await conn.execute(
        """CREATE TABLE IF NOT EXISTS crypto_olm_session (
            account_id     TEXT,
            session_id     CHAR(43),
            sender_key     CHAR(43)  NOT NULL,
            session        bytea     NOT NULL,
            created_at     timestamp NOT NULL,
            last_decrypted timestamp NOT NULL,
            last_encrypted timestamp NOT NULL,
            PRIMARY KEY (account_id, session_id)
        )"""
    )
    await conn.execute(
        """CREATE TABLE IF NOT EXISTS crypto_megolm_inbound_session (
            account_id   TEXT,
            session_id   CHAR(43),
            sender_key   CHAR(43)  NOT NULL,
            signing_key  CHAR(43)  NOT NULL,
            room_id      TEXT      NOT NULL,
            session      bytea     NOT NULL,
            forwarding_chains TEXT NOT NULL,
            PRIMARY KEY (account_id, session_id)
        )"""
    )
    await conn.execute(
        """CREATE TABLE IF NOT EXISTS crypto_megolm_outbound_session (
            account_id    TEXT,
            room_id       TEXT,
            session_id    CHAR(43)  NOT NULL UNIQUE,
            session       bytea     NOT NULL,
            shared        BOOLEAN   NOT NULL,
            max_messages  INTEGER   NOT NULL,
            message_count INTEGER   NOT NULL,
            max_age       INTERVAL  NOT NULL,
            created_at    timestamp NOT NULL,
            last_used     timestamp NOT NULL,
            PRIMARY KEY (account_id, room_id)
        )"""
    )
    await conn.execute(
        """CREATE TABLE crypto_cross_signing_keys (
            user_id TEXT,
            usage   TEXT,
            key     CHAR(43),
            first_seen_key CHAR(43),
            PRIMARY KEY (user_id, usage)
        )"""
    )
    await conn.execute(
        """CREATE TABLE crypto_cross_signing_signatures (
            signed_user_id TEXT,
            signed_key     TEXT,
            signer_user_id TEXT,
            signer_key     TEXT,
            signature      TEXT,
            PRIMARY KEY (signed_user_id, signed_key, signer_user_id, signer_key)
        )"""
    )


@upgrade_table.register(description="Add account_id primary key column")
async def upgrade_v2(conn: Connection, scheme: Scheme) -> None:
    if scheme == Scheme.SQLITE:
        await conn.execute("DROP TABLE crypto_account")
        await conn.execute("DROP TABLE crypto_olm_session")
        await conn.execute("DROP TABLE crypto_megolm_inbound_session")
        await conn.execute("DROP TABLE crypto_megolm_outbound_session")
        await conn.execute(
            """CREATE TABLE crypto_account (
                account_id VARCHAR(255) PRIMARY KEY,
                device_id  VARCHAR(255) NOT NULL,
                shared     BOOLEAN      NOT NULL,
                sync_token TEXT         NOT NULL,
                account    bytea        NOT NULL
            )"""
        )
        await conn.execute(
            """CREATE TABLE crypto_olm_session (
                account_id   VARCHAR(255),
                session_id   CHAR(43),
                sender_key   CHAR(43)  NOT NULL,
                session      bytea     NOT NULL,
                created_at   timestamp NOT NULL,
                last_used    timestamp NOT NULL,
                PRIMARY KEY (account_id, session_id)
            )"""
        )
        await conn.execute(
            """CREATE TABLE crypto_megolm_inbound_session (
                account_id   VARCHAR(255),
                session_id   CHAR(43),
                sender_key   CHAR(43)     NOT NULL,
                signing_key  CHAR(43)     NOT NULL,
                room_id      VARCHAR(255) NOT NULL,
                session      bytea        NOT NULL,
                forwarding_chains TEXT    NOT NULL,
                PRIMARY KEY (account_id, session_id)
            )"""
        )
        await conn.execute(
            """CREATE TABLE crypto_megolm_outbound_session (
                account_id    VARCHAR(255),
                room_id       VARCHAR(255),
                session_id    CHAR(43)     NOT NULL UNIQUE,
                session       bytea        NOT NULL,
                shared        BOOLEAN      NOT NULL,
                max_messages  INTEGER      NOT NULL,
                message_count INTEGER      NOT NULL,
                max_age       INTERVAL     NOT NULL,
                created_at    timestamp    NOT NULL,
                last_used     timestamp    NOT NULL,
                PRIMARY KEY (account_id, room_id)
            )"""
        )
    else:

        async def add_account_id_column(table: str, pkey_columns: list[str]) -> None:
            await conn.execute(f"ALTER TABLE {table} ADD COLUMN account_id VARCHAR(255)")
            await conn.execute(f"UPDATE {table} SET account_id=''")
            await conn.execute(f"ALTER TABLE {table} ALTER COLUMN account_id SET NOT NULL")
            await conn.execute(f"ALTER TABLE {table} DROP CONSTRAINT {table}_pkey")
            pkey_columns.append("account_id")
            pkey_columns_str = ", ".join(f'"{col}"' for col in pkey_columns)
            await conn.execute(
                f"ALTER TABLE {table} ADD CONSTRAINT {table}_pkey "
                f"PRIMARY KEY ({pkey_columns_str})"
            )

        await add_account_id_column("crypto_account", [])
        await add_account_id_column("crypto_olm_session", ["session_id"])
        await add_account_id_column("crypto_megolm_inbound_session", ["session_id"])
        await add_account_id_column("crypto_megolm_outbound_session", ["room_id"])


@upgrade_table.register(description="Stop using size-limited string fields")
async def upgrade_v3(conn: Connection, scheme: Scheme) -> None:
    if scheme == Scheme.SQLITE:
        return
    await conn.execute("ALTER TABLE crypto_account ALTER COLUMN account_id TYPE TEXT")
    await conn.execute("ALTER TABLE crypto_account ALTER COLUMN device_id TYPE TEXT")
    await conn.execute("ALTER TABLE crypto_message_index ALTER COLUMN event_id TYPE TEXT")
    await conn.execute("ALTER TABLE crypto_tracked_user ALTER COLUMN user_id TYPE TEXT")
    await conn.execute("ALTER TABLE crypto_device ALTER COLUMN user_id TYPE TEXT")
    await conn.execute("ALTER TABLE crypto_device ALTER COLUMN device_id TYPE TEXT")
    await conn.execute("ALTER TABLE crypto_device ALTER COLUMN name TYPE TEXT")
    await conn.execute("ALTER TABLE crypto_olm_session ALTER COLUMN account_id TYPE TEXT")
    await conn.execute(
        "ALTER TABLE crypto_megolm_inbound_session ALTER COLUMN account_id TYPE TEXT"
    )
    await conn.execute("ALTER TABLE crypto_megolm_inbound_session ALTER COLUMN room_id TYPE TEXT")
    await conn.execute(
        "ALTER TABLE crypto_megolm_outbound_session ALTER COLUMN account_id TYPE TEXT"
    )
    await conn.execute("ALTER TABLE crypto_megolm_outbound_session ALTER COLUMN room_id TYPE TEXT")


@upgrade_table.register(description="Split last_used into last_encrypted and last_decrypted")
async def upgrade_v4(conn: Connection, scheme: Scheme) -> None:
    await conn.execute("ALTER TABLE crypto_olm_session RENAME COLUMN last_used TO last_decrypted")
    await conn.execute("ALTER TABLE crypto_olm_session ADD COLUMN last_encrypted timestamp")
    await conn.execute("UPDATE crypto_olm_session SET last_encrypted=last_decrypted")
    if scheme == Scheme.POSTGRES:
        # This is too hard to do on sqlite, so let's just do it on postgres
        await conn.execute(
            "ALTER TABLE crypto_olm_session ALTER COLUMN last_encrypted SET NOT NULL"
        )


@upgrade_table.register(description="Add cross-signing key and signature caches")
async def upgrade_v5(conn: Connection) -> None:
    await conn.execute(
        """CREATE TABLE crypto_cross_signing_keys (
            user_id TEXT,
            usage   TEXT,
            key     CHAR(43),
            first_seen_key CHAR(43),
            PRIMARY KEY (user_id, usage)
        )"""
    )
    await conn.execute(
        """CREATE TABLE crypto_cross_signing_signatures (
            signed_user_id TEXT,
            signed_key     TEXT,
            signer_user_id TEXT,
            signer_key     TEXT,
            signature      TEXT,
            PRIMARY KEY (signed_user_id, signed_key, signer_user_id, signer_key)
        )"""
    )


@upgrade_table.register(description="Update trust state values")
async def upgrade_v6(conn: Connection) -> None:
    await conn.execute("UPDATE crypto_device SET trust=300 WHERE trust=1")  # verified
    await conn.execute("UPDATE crypto_device SET trust=-100 WHERE trust=2")  # blacklisted
    await conn.execute("UPDATE crypto_device SET trust=0 WHERE trust=3")  # ignored -> unset



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import pytest

from . import parse_html


async def test_basic_markdown() -> None:
    tests = {
        "<b>test</b>": "**test**",
        "<strong>t<em>e<del>s</del>t</em>!</strong>": "**t_e~~s~~t_!**",
        "<a href='https://example.com'>example</a>": "[example](https://example.com)",
        "<pre><code class='language-css'>div {\n    display: none;\n}</code></pre>": "```css\ndiv {\n    display: none;\n}\n```",
        "<code>hello</code>": "`hello`",
        "<blockquote>Testing<br>123</blockquote>": "> Testing\n> 123",
        "<ul><li>test</li>\n<li>foo</li>\n<li>bar</li>\n</ul>": "● test\n● foo\n● bar",
        "<ol start=123><li>test</li>\n<li>foo</li>\n<li>bar</li>\n</ol>": "123. test\n124. foo\n125. bar",
        "<h4>header</h4>": "#### header",
        "<span data-mx-spoiler>spoiler?</span>": "||spoiler?||",
        "<span data-mx-spoiler='SPOILER!'>not really</span>": "||SPOILER!|not really||",
    }
    for html, markdown_ish in tests.items():
        assert await parse_html(html) == markdown_ish


async def test_nested_markdown() -> None:
    input_html = """
<h1>Hello, World!</h1>
<blockquote>
  <ol>
    <li><a href='https://example.com'>example</a></li>
    <li>
      <ul>
        <li>item 1</li>
        <li>item 2</li>
      </ul>
    </li>
    <li>
      <pre><code class="language-python">def random() -> int:
    if 4 is 1:
        return 5
    return 4</code></pre>
    </li>
    <li><strong>Just some text</strong></li>
  </ol>
</blockquote>
""".strip()
    expected_output = """
# Hello, World!
> 1. [example](https://example.com)
> 2. ● item 1
>    ● item 2
> 3. ```python
>    def random() -> int:
>        if 4 is 1:
>            return 5
>        return 4
>    ```
> 4. **Just some text**
""".strip()
    assert await parse_html(input_html) == expected_output
